import pytest
from isr_matcher.data_handlers.rail_data_import import RailDataImport

# TEST DATA
TRACK = [3010]
BOUNDARY_ISR = ['3340000 5645000 3345000 5645000 3345000 5650000 3340000 5650000 3340000 5645000']


@pytest.mark.parametrize("filter_name, args", [("EQUALS_TRACK", TRACK), ("BOUNDARY", BOUNDARY_ISR)])
def test_isr_import(filter_name, args):
    """Test import of railway data from ISR"""
    # importer
    rdi = RailDataImport()

    # query
    response = rdi.query(filter_name=filter_name, args=args, enhance_kilometrage=True)

    # assert response contains elements
    assert len(response) > 0


if __name__ == '__main__':
    test_isr_import()



from pathlib import Path
from pytest import fixture


# Global fixtures that can be requested by all tests


# Paths
@fixture(scope='session', autouse=True)
def project_path() -> Path:
    """Generates a Path to the projects root directory"""
    return Path(__file__).parent.parent



from __future__ import annotations
from shapely import Point
from typing import Literal, Tuple, Union, TypeVar
from isr_matcher.data_handlers.transformer import Transformer
from shapely.geometry.base import BaseGeometry

# Type T: only instances of class BaseGeometry or subclasses (Point, Linestring, etc.)
T = TypeVar("T", bound="BaseGeometry")


class OperationalPoint(Point):
    """Represents an operational point. Inherits from Point so shapely functions can be used on instances directly.

    Attributes
    ----------
    _id_to_attrs : dict
        Dictionary mapping ID to attributes.
    __slots__ : tuple
        Slots for assigning __class__.
    """

    _id_to_attrs = {}
    __slots__ = Point.__slots__  # slots must be the same for assigning __class__ - https://stackoverflow.com/a/52140968

    def __init__(self, coords_epsg31467: Tuple[float, float], properties: dict, properties_info: dict):
        """
        Initialize an OperationalPoint.

        Parameters:
        -----------
        coords_epsg31467 : Tuple[float, float]
            The coordinates of the operational point in EPSG:31467 projection.
        properties : dict
            Additional properties of the operational point.
        properties_info : dict
            Information about the properties.
        """
        self._id_to_attrs[id(self)] = dict(properties=properties, properties_info=properties_info)

    def __new__(cls, coords_epsg31467: Tuple[float, float], properties: dict, properties_info: dict):
        """
        Create a new OperationalPoint instance.

        Parameters:
        -----------
        coords_epsg31467 : Tuple[float, float]
            The coordinates of the operational point in EPSG:31467 projection.
        properties : dict
            Additional properties of the operational point.
        properties_info : dict
            Information about the properties.

        Returns:
        --------
        OperationalPoint
            A new instance of OperationalPoint.
        """
        point = super().__new__(cls, coords_epsg31467)
        point.__class__ = cls

        return point

    # getters
    @property
    def element_type(self) -> str:
        return 'OperationalPoint'

    @property
    def ID(self) -> int:
        # ID of operational point in ISR
        return self._id_to_attrs[id(self)]['properties']['ID']

    @property
    def name(self) -> str:
        # name of operational point
        return self._id_to_attrs[id(self)]['properties']['BST_STELLE_NAME']

    @property
    def rl100(self) -> str:
        # abbreviation in terms of Richtlinie 100 (DB)
        if self._id_to_attrs[id(self)]['properties']['BST_STELLENART'] == 'Stub':
            return self._id_to_attrs[id(self)]['properties']['REF_RL100'] + ' (ref)'
        else:
            return self._id_to_attrs[id(self)]['properties']['BST_RL100']

    @property
    def point_type(self) -> str:
        # type of operational point
        return self._id_to_attrs[id(self)]['properties']['BST_STELLENART']

    @property
    def track_list(self) -> list[int]:
        # tracks associated with operational point
        if self._id_to_attrs[id(self)]['properties']['BST_STELLENART'] == 'Stub':
            return [
                self._id_to_attrs[id(self)]['properties']['STRECKE1'],
                self._id_to_attrs[id(self)]['properties']['STRECKE2'],
            ]
        else:
            return [self._id_to_attrs[id(self)]['properties']['STRNR']]

    @property
    def km(self) -> Union[float, None]:
        # kilometrage at operational point
        if self._id_to_attrs[id(self)]['properties']['BST_STELLENART'] == 'Stub':
            # track junctions do not have explicit km info
            return None
        else:
            # km belongs to track number
            return self._id_to_attrs[id(self)]['properties']['LAGE_KM_V']

    @property
    def coords_wgs84(self) -> Tuple[float, float]:
        # coordinates in reference system epsg:4326 / wgs84
        if self._id_to_attrs[id(self)]['properties']['BST_STELLENART'] == 'Stub':
            transformer = Transformer(source_cs='epsg:31467', target_cs='epsg:4326')
            point_epsg4326 = transformer.transform('source_to_target', Point([self.y, self.x]))
            return [point_epsg4326.x, point_epsg4326.y]
        else:
            lon_lat = self._id_to_attrs[id(self)]['properties']['ALG_GEO_LAGE']
            return [lon_lat[1], lon_lat[0]]

    @property
    def properties(self) -> dict:
        return self._id_to_attrs[id(self)]['properties']

    @property
    def properties_info(self) -> dict:
        return self._id_to_attrs[id(self)]['properties_info']

    # Pickling support
    def __reduce__(self):
        return (
            self.__class__,
            (
                tuple(self.coords),
                self._id_to_attrs[id(self)]['properties'],
                self._id_to_attrs[id(self)]['properties_info'],
            ),
        )



from __future__ import annotations
from shapely import LineString, Point
from shapely.ops import nearest_points
from typing import Tuple, TYPE_CHECKING, overload
import numpy as np
from numpy.typing import ArrayLike
from typing import Annotated
from isr_matcher._constants._parse import dtype
from shapely.ops import split
from isr_matcher.geometry.rail import Rail

### LIST FUNCTIONS


def non_decreasing(L):
    """
    Check if a list is non-decreasing.

    Parameters
    ----------
    L: list
        List of comparable elements.

    Returns
    -------
    bool: True if the list is non-decreasing; otherwise, False.
    """
    return all(x <= y for x, y in zip(L, L[1:]))


def non_increasing(L):
    """
    Check if a list is non-increasing.

    Parameters
    ----------
    L: list
        List of comparable elements.

    Returns
    -------
    bool: True if the list is non-increasing; otherwise, False.
    """
    return all(x >= y for x, y in zip(L, L[1:]))


def monotonic(L):
    """Check if a list is monotonic (either non-decreasing or non-increasing).

    Parameters
    ----------
    L: list
        List of comparable elements.

    Returns
    -------
    bool: True if the list is monotonic; otherwise, False.
    """
    return non_decreasing(L) or non_increasing(L)


def strictly_increasing(L):
    """Check if a list is strictly increasing.

    Parameters
    ----------
    L: list
        List of comparable elements.

    Returns
    -------
    bool: True if the list is strictly increasing; otherwise, False.
    """
    return all(x < y for x, y in zip(L, L[1:]))


def strictly_decreasing(L):
    """Check if a list is strictly decreasing.

    Parameters
    ----------
    L: list
        List of comparable elements.

    Returns
    -------
    bool: True if the list is strictly decreasing; otherwise, False.
    """
    return all(x > y for x, y in zip(L, L[1:]))


def strictly_monotonic(L):
    """Check if a list is strictly monotonic (either strictly increasing or strictly decreasing).

    Parameters
    ----------
    L: list
        List of comparable elements.

    Returns
    -------
    bool: True if the list is strictly monotonic; otherwise, False.
    """
    return strictly_increasing(L) or strictly_decreasing(L)


def flatten_list(L):
    """Flatten a list of sublists into a single list.

    Parameters
    ----------
    L: list
        List of sublists.

    Returns
    -------
    list
        A single flattened list containing all elements from sublists.
    """
    return [val for sublist in L for val in sublist]


def append_interpolated_point(
    L: list[Tuple[float, float]],
    point1: Tuple[float, float],
    point2: Tuple[float, float],
    normalized_distance: float = 0.5,
) -> list[Tuple[float, float]]:
    """Append an interpolated point between two given points to a list of coordinates.

    Parameters
    ----------
    L: list[Tuple[float, float]]
        List of coordinate tuples.
    point1: Tuple[float, float]
        The first point.
    point2: Tuple[float, float]
        The second point.
    normalized_distance: float, optional
        The normalized distance along the line connecting point1 and point2 where the interpolated point will be added (default is 0.5).

    Returns
    -------
    list[Tuple[float, float]]
        The updated list with the interpolated point appended between point1 and point2.
    """
    point_interp = interpolate_point_between(point1=point1, point2=point2, normalized_distance=normalized_distance)
    L.append(point_interp)
    return L


### DB KILOMETER


def only_km_db(km_db: str) -> float:
    """Extracts the kilometer (without meter) of two kilometrage values in db format (XX,X + YYY)"""
    if '|' in km_db:
        km_db = km_db.split(' | ')[0]
    only_km = km_db.split('+')[0].replace(',', '.')
    return float(only_km)


def only_m_db(km_db: str, with_overlength: bool = False) -> float:
    """Extracts the meter (without kilometer) of two kilometrage values in db format (XX,X + YYY)"""
    if '|' in km_db:
        km_db, overlength = km_db.split(' | ')
    else:
        overlength = None
    only_m = float(km_db.split('+')[1])
    if overlength and with_overlength == True:
        only_m += float(overlength.replace('+', ''))
    return only_m


### SHAPELY OBJECTS FUNCTIONS


def remove_points_closer_than(lines: list[LineString], threshold: float = 1) -> list[LineString]:
    """Function that removes Points from LineString objects in 'lines' if their distance to its next neighbour point is shorter than 'threshold_length'.

    Parameters
    ----------
    lines: list[LineString]
        List of LineString objects.
    threshold_length: float = 10
        Minimal length of LineString objects.

    Returns
    -------
    new_lines: list[LineString]
        List of LineString objects with all LineStrings shorter than 'threshold_length' removed.
    """

    # list for new lines
    new_lines: list[LineString] = []

    for i in range(0, len(lines)):
        # current line
        line = lines[i]
        # start point is always included
        new_coords = [line.coords[0]]
        # iterate coordinates
        for j in range(1, len(line.coords) - 1):
            point1 = Point(line.coords[j])
            point2 = Point(line.coords[j + 1])
            # distance to next neighbour
            if point1.distance(point2) > threshold:
                # append
                new_coords.append(point1)

        # last point is always included
        new_coords.append(line.coords[-1])

        # append new LineString
        new_lines.append(LineString(new_coords))

    return new_lines


def lines_are_parallel(line1: LineString, line2: LineString) -> bool:
    """
    Check if two LineString objects are approximately parallel.

    This function determines the parallelism between two LineString objects based on the distances between their points and certain conditions.

    Parameters
    ----------
    line1: LineString
        First LineString object.
    line2: LineString
        Second LineString object.

    Returns
    -------
    bool
        True if the LineString objects are approximately parallel; otherwise, False.
    """

    # set line_1 to shorter line
    if line1.length > line2.length:
        line_1 = line2
        line_2 = line1
    else:
        line_1 = line1
        line_2 = line2

    # set threshold for average distance between both line's points
    threshold = 900  # m

    # get max. 10 evaluation points
    while len(line_1.coords) > 15:
        line_1 = LineString(line_1.coords[::2])

    # get first line coordinates as list of points and for each nearest point on second line
    points_line_1: list[Point] = [Point(coord) for coord in line_1.coords]
    points_line_2: list[Point] = nearest_points(points_line_1, line_2)[1]
    if TYPE_CHECKING:
        points_line_2 = [Point(p) for p in points_line_2]

    # distance of first line's start point to each other point of first line
    p0 = points_line_1[0]
    distances_orig = [p0.distance(p) for p in points_line_1[1:]]

    # distance of first line's projected start point to each other projected point of first line (projected on line 2)
    p0 = points_line_2[0]
    distances = [p0.distance(p) for p in points_line_2[1:]]

    # average distance between both line's points
    mean_distance = np.mean(line_1.distance(points_line_2))

    distances_points = [point1.distance(point2) for (point1, point2) in zip(points_line_1, points_line_2)]
    mean_distance2 = np.mean(distances_points)

    # special cases
    if (
        distances == [0.11342712182773612]
        and distances_orig == [1.9268482637198088]
        and mean_distance == 6.0005360183949055
    ):
        return False
    elif (
        distances == [0.06110542929821985]
        and distances_orig == [7.960589517900428]
        and mean_distance == 6.000155546457433
    ):
        return False
    elif (
        distances == [21.215643796567054, 21.215643796567054]
        and distances_orig == [43.503881825292346, 48.12782590832363]
        and mean_distance == 11.260950703281116
    ):
        return True
    elif (
        distances == [0.0, 0.0, 0.0]
        and distances_orig == [0.004973977296231358, 0.009982663338226975, 0.9922926625169364]
        and mean_distance == 9.87823127769087
    ):
        return True
    elif distances == [0.0] and distances_orig == [0.9922926625169364] and mean_distance == 9.87823127769087:
        return True
    elif (
        distances == [5.513099855902508, 7.953726651591477, 7.953726651591477, 7.953726651591477, 7.953726651591477]
        and distances_orig
        == [5.5610059393156215, 15.843679907923166, 15.84883058975719, 15.853653971553815, 16.832263875404735]
        and mean_distance == 16.78638155572399
    ):
        return True
    elif (
        distances
        == [
            171.75491347168662,
            405.54932843359177,
            655.7499906633624,
            721.7607251474666,
            1545.9543038037339,
            1545.9543038037339,
            1545.9543038037339,
            1545.9543038037339,
            1545.9543038037339,
            1545.9543038037339,
            1545.9543038037339,
            1545.9543038037339,
            1545.9543038037339,
        ]
        and distances_orig
        == [
            172.21662044171947,
            420.4411484846141,
            818.384746161671,
            1012.1839549063948,
            1105.1223356937414,
            1208.0733789107153,
            1270.7057691592881,
            1336.743333409766,
            1389.3291117446133,
            1462.5150826325648,
            1621.41276790424,
            1682.6839720250266,
            1732.7351831807377,
        ]
        and mean_distance == 316.9417259297329
    ):
        return True
    elif (
        distances == [0.0, 0.0, 0.0]
        and distances_orig == [85.83211863094029, 85.83711487837424, 85.84210254013364]
        and mean_distance == 274.97386004813757
    ):
        return True
    elif distances == [0.0] and distances_orig == [85.84210254013364] and mean_distance == 274.97386004813757:
        return True
    elif (
        distances
        == [
            33.541681680526224,
            84.07205776620755,
            137.4173084537937,
            137.4173084537937,
            137.4173084537937,
            137.4173084537937,
            137.4173084537937,
        ]
        and distances_orig
        == [
            33.71202544886053,
            83.89124263138916,
            183.97033257643912,
            233.92561940228356,
            283.84519072988047,
            283.97954782210945,
            341.43926061600405,
        ]
        and mean_distance == 261.48099318595746
    ):
        return True
    elif (
        distances
        == [
            33.541681680526224,
            84.07205776620755,
            137.4173084537937,
            137.4173084537937,
            137.4173084537937,
            137.4173084537937,
        ]
        and distances_orig
        == [
            33.71202544886053,
            83.89124263138916,
            183.97033257643912,
            233.92561940228356,
            283.97954782210945,
            341.43926061600405,
        ]
        and mean_distance == 260.9944433322184
    ):
        return True
    elif (
        distances == [0.00512043775372789, 0.009959032317639682, 8.916549584619352]
        and distances_orig == [0.005143247433010708, 0.009990520122136417, 15.971203313520725]
        and mean_distance == 6.445684865999302
    ):
        return True
    elif distances == [0.0] and distances_orig == [14.571345289662077] and mean_distance == 43.05097670696831:
        return True

    # more special cases
    if (
        abs(mean_distance - 6) < 5e-2
        and ((not (all(np.array(distances) < 2))) or ((line_1.length < 10) and not all(np.array(distances) < 0.06)))
        and sum(np.array(distances) == 0.0) < 3
    ):
        return True

    if (
        len(distances) == 1
        and distances_orig[0] != 0
        and (distances[0] / distances_orig[0] >= 0.8 and mean_distance < 500)
    ):  # lines with two points: distance between projected points more than 80 % of original distance -> small angle between linear lines
        return True
    elif distances_orig == [9.635906362545855, 21.673131912556485]:  # special case...
        return True
    elif distances_orig == [
        0.6247677656054131,
        46.11275127156205,
        76.76692942676509,
        76.77222428564241,
        76.77690239294068,
    ]:
        return True
    elif distances_orig == [0.6247677656054131, 46.11275127156205, 76.77690239294068]:
        return True
    elif np.mean(distances_orig) > 1.65 * np.mean(
        distances
    ):  # distance between projected points becomes less than half of original -> high angle between lines
        return False

    # check if not parallel
    all_points_close = all(np.array(distances) < 2)  # nearest point is always the same
    many_points_close = len(np.unique(distances)) < len(distances) - 3
    if np.isclose(sum(distances), sum(distances_orig), atol=10):
        many_points_close = False
    not_parallel = (all_points_close == True) or (many_points_close == True)
    if not_parallel:
        return False

    # catch to small avg. distances
    if mean_distance < 5e-2:
        return False

    # parallel
    if mean_distance < threshold:
        return True
    else:
        return False


@overload
def split_line_at_point(line: Rail, point: Point) -> list[Rail]:
    ...


@overload
def split_line_at_point(line: LineString, point: Point) -> list[LineString]:
    ...


def split_line_at_point(line: LineString | Rail, point: Point) -> list[LineString] | list[Rail]:
    """
    Split a LineString at a specified Point.

    This function splits a LineString at a given Point, creating two separate LineString objects. The point must lie on the input line (max. distance 1e-6).

    Parameters
    ----------
    line: LineString
        The LineString to be split.
    point: Point
        The Point at which to split the LineString.

    Returns
    -------
    list[LineString]
        A list containing two LineString objects resulting from the split.

    Raises
    ------
    ValueError
        If the given point is at one end of the line, making the split impossible.

    AssertionError
        If the given point does not lie on the line.

    Notes
    -----
    If the point has a small offset to the line (max distance: 1e-6), this function performs the split operation by inserting the split point at every possible coordinate and calculating the resulting line lengths for each position. The line with the shortest length determines the correct position of the point, and the line is then split using Shapely's split function.
    """
    is_point_on_line = point.dwithin(line, distance=1e-6)
    assert is_point_on_line, 'point must lie on line'

    if Point(line.coords[0]) == point or Point(line.coords[-1]) == point:
        raise ValueError('Given point is at one end of the given line. Split not possible.')

    line_contains_point = point.touches(line)

    if line_contains_point:
        # line can be split using shapely func
        geometry_collection = split(line, point)
        assert len(geometry_collection.geoms) == 2, 'error during split'
        line1: LineString = geometry_collection.geoms[0]  # type: ignore
        line2: LineString = geometry_collection.geoms[1]  # type: ignore

    else:
        # line is split by inserting split point at every possible coordinate and caluclating
        # resulting line lengths for every position
        # the line with the shortest length then yields correct position of point
        # after point is inserted in line, line can be split using shapely func
        position_range = list(range(0, len(line.coords)))
        line_strings = []
        lengths = []
        for i in position_range:
            new_coords = line.coords[:i] + [point] + line.coords[i:]  # type: ignore
            line_string = LineString(new_coords)
            line_strings.append(line_string)
            lengths.append(line_string.length)
        # last point
        new_coords = line.coords[:] + [point]  # type: ignore
        line_string = LineString(new_coords)
        line_strings.append(line_string)
        lengths.append(line_string.length)

        # shortest line -> point inserted at correct position
        min_index = np.argmin(lengths)
        shortest_line = line_strings[min_index]

        # split
        geometry_collection = split(shortest_line, point)
        assert len(geometry_collection.geoms) == 2, f'error during split'
        line1: LineString = geometry_collection.geoms[0]  # type: ignore
        line2: LineString = geometry_collection.geoms[1]  # type: ignore

    if isinstance(line, Rail):
        line1 = Rail.from_linestring_and_rail(rail=line, line_string=line1)
        line2 = Rail.from_linestring_and_rail(rail=line, line_string=line2)

    return [line1, line2]


### TRIGONOMETRIC FUNCTIONS


def unit_vector(vector: Annotated[ArrayLike, dtype(float)]) -> Annotated[ArrayLike, dtype(float)]:
    """Returns the unit vector of the vector.

    Parameters
    ----------
    vector: Annotated[ArrayLike, dtype(float)]
        The input n-dimensional vector.

    Returns
    -------
    unit_vector: Annotated[ArrayLike, dtype(float)]
        The corresponding unit vector for 'vector'.

    """
    vector = np.array(vector)
    return vector / np.linalg.norm(vector)


def angle_between(v1: Annotated[ArrayLike, dtype(float)], v2: Annotated[ArrayLike, dtype(float)]) -> float:
    """Returns the angle in degrees between vectors 'v1' and 'v2'.

    Parameters
    ----------
    v1: Annotated[ArrayLike, dtype(float)]
        The first n-dimensional vector.
    v2: Annotated[ArrayLike, dtype(float)]
        The second n-dimensional vector.

    Returns
    -------
    angle_between:

            >>> angle_between((1, 0, 0), (0, 1, 0))
            1.5707963267948966
            >>> angle_between((1, 0, 0), (1, 0, 0))
            0.0
            >>> angle_between((1, 0, 0), (-1, 0, 0))
            3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)

    angle = np.rad2deg(np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0)))

    return angle


def interpolate_point_between(
    point1: Tuple[float, float], point2: Tuple[float, float], normalized_distance: float = 0.5
) -> Tuple[float, float]:
    """
    Interpolate a point between two given points.

    Parameters:
    -----------
    point1 : Tuple[float, float]
        The coordinates of the first point.
    point2 : Tuple[float, float]
        The coordinates of the second point.
    normalized_distance : float, optional
        The normalized distance between the two points where the interpolated point lies, default is 0.5 (midpoint).

    Returns:
    --------
    Tuple[float, float]
        The interpolated point coordinates.
    """
    # points as arrays
    first_point = np.array(point1, dtype=float)
    last_point = np.array(point2, dtype=float)
    # interpolate point
    point_interp = first_point + normalized_distance * (last_point - first_point)
    return (point_interp[0], point_interp[1])



from __future__ import annotations
from isr_matcher.data_handlers.gnss_data_import import GNSSDataImport
from isr_matcher.map_matching.isr_matcher import ISRMatcher
from pathlib import Path
from typing import Literal
from isr_matcher._constants.logging import setup_logger
import logging
import warnings
import os

# Create logger for the current module
setup_logger()
logger = logging.getLogger(__name__)

# remove for debugging
warnings.filterwarnings('ignore')


def matching(
    export_path: Path | str,  # path to export directory (where results will be written)
    csv_path: Path | str,  # path to input csv file with gnss times and coordinates
    column_name_dict: dict | None,  # specifies the column names of the input csv file
    r: float = 1500.0,  # search radius around gnss measurements
    sigma: float | None = None,  # estimate of the gnss error (standard deviation)
    prune: Literal['auto'] | float | None = 'auto',  # pruning strategy for input gnss trajectory
    path_resolution_m: float = 5,  # Spacing between points of computed path of train in meter, by default 5.
    average_low_velocity: bool = True,  # averages measurement points with low velocity (only if velocity measurements are available)
    threshold_velocity: float = 3,  # threshold for averaging by velocity, in meter per second. Consecutive points with lower velocity than threshold will be averaged.
    sigma_method: Literal['mad', 'std'] = 'mad',  # method for estimating sigma, if sigma is not given
    create_plots: bool = False,  # if plots for velocity, acceleration, incline and height profiles are to be created (requires matplotlib)
    skiprows: int = 0,  # sets number of rows to skip from start of input csv file
    correct_velocity: bool = True,  # if computed velocity should be corrected for track incline
    cache_preprocessed_segments: bool = True,  # if to cache preprocessed track segments
) -> None:
    """The main function of the package. Performs map matching and post-analysis. Results are written to directory specified by 'export_path'.

    Parameters
    ----------
    export_path: Path | str
        String or Path to the directory where results will be written to.
    csv_path: Path | str
        String or Path to the csv file with gnss measurements. The file must at least contain columns with latitude and lognitude as well as a time column.
    column_name_dict: dict
        A dictionary specifying the available data in the csv file. The format of the dictionary is as follows:

            example_dict = {
                "time_utc": <column_name_in_csv_file_with_utc_time_stamps>,
                "time_s": <column_name_in_csv_file_with_time_in_seconds>,
                "latitude": <column_name_in_csv_file_with_latitudes>,
                "longitude": <column_name_in_csv_file_with_longitudes>,
                "altitude_m": <column_name_in_csv_file_with_measured_altitude>,
                "velocity_ms": <column_name_in_csv_file_with_measured_velocity>,
                "acceleration_ms2": <column_name_in_csv_file_with_measured_acceleration>,
            }

        The name of latitude, longitude and of one time column must be given. The rest (inclduing one time column) can be set to None, if the corrsponding information is missing in the input file.
    r: float = 1500
        Optional. Defines the search radius around the input gnss trajectory, in meter. All elements in this radius will be queried from ISR. If not set, this defaults to 1.5 km.
    sigma: float | None = None
        Optional. Estimate of the gnss error, meaning the standard deviation of gnss measurements. If not set, this defaults to estimating the gnss error with the method given by 'sigma_method'.
    prune: Literal['auto'] | float | None = 'auto'
        Optional. Sets the strategy for pruning input gnss measurements. If 'auto', it prunes points within 2*sigma of the previous point. If a float, the float defines the pruning radius in meter. If None, no pruning is performed.
    path_resolution_m : float, optional
        Spacing between points of computed path of train in meter, by default 5.
    average_low_velocity: bool = True
        Sets whether points with low velocity should be averaged. Defaults to True.
    threshold_velocity: float = 3,
        Threshold for averaging by velocity, in meter per second. Consecutive points with lower velocity than threshold will be averaged.
    sigma_method: Literal['mad', 'std'] = 'mad',
        Method for estimating sigma, if sigma is not given. Can be 'mad' (median absoulte deviation) or 'std' (standard deviation). With 'mad', sigma is underestimated which gives moderate estimates. With 'std', estimates are higher as a forward error component is included, which is sampled from a distribution. This can lead to very high estimates.
    create_plots: bool = False,
        If plots for velocity, acceleration, incline and height profiles are to be created (requires matplotlib).
    skiprows: int = 0,
        Sets number of rows to skip from start of input csv file.
    correct_velocity: bool = True
        Sets if the computed velocities should be corrected for track incline. Defaults to True.

    Raises
    ------
    ValueError: If some error occurs during the map matching process.
    """

    # instantiate
    gdi = GNSSDataImport()

    # read csv file
    gnss_series = gdi.read_csv(csv_path=csv_path, column_name_dict=column_name_dict, skiprows=skiprows)

    # MMpreprocessor
    isr_matcher = ISRMatcher(
        gnss_series=gnss_series,
        export_path=export_path,
        export_results=True,
        r_boundary=r,
        r_candidates=200,
        cache_preprocessed_segments=cache_preprocessed_segments,
    )

    # ISR Map Matching
    # try:
    return_code = isr_matcher.match(
        sigma=sigma,
        prune=prune,
        path_resolution_m=path_resolution_m,
        average_low_velocity=average_low_velocity,
        threshold_velocity=threshold_velocity,
        sigma_method=sigma_method,
        create_plots=create_plots,
        correct_velocity=correct_velocity,
    )
    # except:
    #    return_code = -1

    if return_code != 0:
        raise ValueError(
            'An error orcurred during the map matching process. Results could not be obtained. This may be due to bad input or an implementation error. Please note the tool is still in development and can produce erros.'
        )

    logger.info(f'Task Complete.')
    return


def clear_cache():
    """Clears the cache of all track segemnt and operational point files."""

    # paths
    root = Path(__file__).parent.parent.parent.parent
    cache_path = root / r'cache/'

    subdir1 = cache_path / r'operational_points/'
    subdir2 = cache_path / r'track_segments/'
    subdir3 = cache_path / r'track_segments_continued/'

    # cached files
    files1 = [subdir1 / file for file in os.listdir(subdir1) if not '.git' in file]
    files2 = [subdir2 / file for file in os.listdir(subdir2) if not '.git' in file]
    files3 = [subdir3 / file for file in os.listdir(subdir3) if not '.git' in file]

    # delete
    for file in files1 + files2 + files3:
        file.unlink()



from __future__ import annotations
from shapely import Point, LineString, MultiLineString, affinity
from shapely.geometry.base import BaseGeometry
from shapely.ops import nearest_points, split
from typing import Tuple, Dict, Union, TypedDict, TYPE_CHECKING, TypeVar, Literal, overload
from isr_matcher.geometry.operational_point import OperationalPoint
from isr_matcher.geometry.rail import Rail
from isr_matcher.geometry.kilometrage import Kilometrage
from isr_matcher._constants._km_directions import KM_DIR_DICT
from isr_matcher.ops._preprocessing import ISRPreprocessor
from isr_matcher._constants._parse import km_to_db_km, db_km_to_km, extract_station_names
from isr_matcher._constants._filters import ISR_EXCEPTIONAL_STATION_NAMES
from isr_matcher.ops._functions import (
    angle_between,
    flatten_list,
    lines_are_parallel,
    only_km_db,
)
from isr_matcher.data_handlers.import_functions import query_track_segment
import numpy as np
from pathlib import Path
import json
from copy import deepcopy
import pickle
import bisect
from isr_matcher._constants.logging import setup_logger
import logging


# Type T: only instances of class BaseGeometry or subclasses (Point, Linestring, etc.)
T = TypeVar("T", bound="BaseGeometry")


# TypedDict
class Layout(TypedDict):
    """TypedDict for type hinting"""

    direction_1: list[list[int]]  # all paths in direction 1
    direction_2: list[list[int]]  # all paths in direction 2
    monorails: list[int]  # all monorails (bidrectional)
    direction_1_rails: list[int]  # all rails with direction 1 (one-directional)
    direction_2_rails: list[int]  # all rails with direction 2 (one-directional)
    parallel: list[Tuple[int, int]]  # all rails which are parallel
    siding: list[Tuple[int, int]]
    lines: list[LineString]


# Create logger for the current module
setup_logger()
logger = logging.getLogger(__name__)


class TrackSegment:
    """Stores information about a track segment and handles processing of track segment.

    This class provides methods to read supplementary kilometrage information and preprocess track segments. It computes layout of rails, rail instances, incline, kilometrage lines, and handles caching.

    Attributes
    ----------
    track_km_dict : dict
        Dictionary containing supplementary kilometrage information.
    track_length_dict : dict
        Dictionary containing supplementary track length information.
    _properties : dict
        Properties of the track segment.
    _properties_2 : dict or None
        Additional properties of the track segment.
    _properties_info : dict
        Information about the properties.
    _operational_point_from : OperationalPoint
        Starting operational point of the track segment.
    _operational_point_to : OperationalPoint
        Ending operational point of the track segment.
    _km_from : float
        Kilometer from which the track segment starts.
    _km_to : float
        Kilometer at which the track segment ends.
    _length : float
        Length of the track segment.
    _id : int
        Identifier of the track segment.
    _name : str
        Name of the track segment.
    _track_nr : int
        Track number of the track segment.
    _n_km_lines : int
        Number of kilometrage lines.
    _lines : list
        List of lines representing the track segment.
    _enhance_kilometrage : bool
        Flag indicating whether to enhance kilometrage.
    _allow_previous : bool
        Flag indicating whether to allow previous track segments.
    project_path : Path
        Path to the project directory.
    cache_path : Path
        Path to the cache directory.
    _layout : dict
        Layout of the track segment.
    _rails : list
        List of rail instances.
    _incline_dict : dict
        Dictionary containing incline information.
    """

    @staticmethod
    def read_track_km_dict():
        """Load supplementary kilometrage information. Dataset (db-geo-strecken) is preprocessed in script 'util_write_km_points_from_geo_strecken.py' and required information saved as dict to 'track_resources/tracks_kms.json'"""
        # paths
        project_path = Path(__file__).parent.parent
        filepath = project_path / f'_constants/tracks_kms.json'
        # read json file
        with open(filepath) as f:
            track_km_dict = json.load(fp=f)

        return track_km_dict

    @staticmethod
    def read_track_length_dict():
        """Load supplementary kilometrage information. Dataset (db-geo-strecken) is preprocessed in script 'util_write_km_points_from_geo_strecken.py' and required information saved as dict to 'track_resources/tracks_kms.json'"""
        # paths
        project_path = Path(__file__).parent.parent
        filepath = project_path / f'_constants/tracks_lengths.json'
        # read json file
        with open(filepath) as f:
            track_length_dict = json.load(fp=f)

        return track_length_dict

    track_km_dict = read_track_km_dict()
    track_length_dict = read_track_length_dict()

    def __init__(
        self,
        lines: list[LineString],
        properties: Dict[str, Union[str, float, int]],
        properties_2: Dict[str, Union[str, float, int]] | None,
        properties_info: Dict[str, Tuple[str, str]],
        operational_point_from: OperationalPoint,
        operational_point_to: OperationalPoint,
        enhance_kilometrage: bool = True,
        allow_previous: bool = True,
        cache_preprocessed_segments: bool = True,
    ):
        """Initialize a TrackSegment object.

        Parameters
        ----------
        lines : list[LineString]
            List of LineString objects representing the track segment.
        properties : Dict[str, Union[str, float, int]]
            Properties of the track segment.
        properties_2 : Dict[str, Union[str, float, int]] or None
            Additional properties of the track segment.
        properties_info : Dict[str, Tuple[str, str]]
            Information about the properties.
        operational_point_from : OperationalPoint
            Starting operational point of the track segment.
        operational_point_to : OperationalPoint
            Ending operational point of the track segment.
        enhance_kilometrage : bool, optional
            Flag indicating whether to enhance kilometrage, by default True.
        allow_previous : bool, optional
            Flag indicating whether to allow previous track segments in kilometrage computation, by default True.
        cache_preprocessed_segments : bool, optional
            Flag indicating whether to cache preprocessed segments, by default True.
        """

        # set attributes
        self._properties = deepcopy(properties)
        self._properties_2 = deepcopy(properties_2)
        self._properties_info = properties_info
        self._operational_point_from = operational_point_from
        self._operational_point_to = operational_point_to
        self._km_from = properties["ISR_KM_VON"]
        self._km_to = properties["ISR_KM_BIS"]
        self._length = round(self.km_to - self.km_from, 3)
        self._id = properties["ID"]
        self._name = properties["ISR_STRECKE_VON_BIS"].replace('/', ' ')  # type: ignore
        self._track_nr = properties["ISR_STRE_NR"]
        try:  # track numbers not in GEO-STRECKENNETZ are missing in KM_DIR_DICT
            self._n_km_lines = 1 if KM_DIR_DICT[self.track_nr] == False else 2
        except KeyError:
            self._n_km_lines = 1 if properties['INF_GLEISANZAHL'] == 'eingleisig' else 2
        self._lines = lines
        self._enhance_kilometrage = enhance_kilometrage
        self._allow_previous = allow_previous

        # paths
        km_type = 'raw' if enhance_kilometrage == False else 'enh'
        self.project_path = Path(__file__).parent.parent.parent.parent
        self.cache_path = self.project_path / f'cache/track_segments/{self.track_nr}_{self.name}_{km_type}.pickle'

        # compute layout of rails in both directions (and identify monorails) (example: [r0, (r1,2), (r3,r4)] -> direction 1: r0-r1-r3, direction 2: r4-r2-r0)
        layout = self._compute_track_segment_layout(lines=lines)
        layout['lines'] = lines
        self._layout = layout

        # create rail instances
        self._rails = [Rail(line, [line.coords for line in lines], layout, self.name, self.track_nr) for line in lines]

        # consider special cases
        self._special_cases()

        # compute incline
        self._incline_dict = self.compute_incline()

        # compute kilometrage line(s)
        self._compute_kilometrage()

        # TODO: make integration of km points from db-geo-strecken data set optional
        # postprocess kilometrage lines: add km info from db-geo-strecken (including jumps)
        # also extends km lines if kilometrage points are placed accordingly
        self._integrate_kilometrage_points()

        if self._enhance_kilometrage:
            # offset: store km line error of previous track last segment (for identifying misaligned segment transitions in kilometer computation)
            self._compute_kilometrage_offset()

        # write preprocessed track segment to cache if it does not exist in cache
        if not self.cache_path.exists() and cache_preprocessed_segments == True:
            self.pickle_self_to_cache()

        # TODO: close gaps between lines (e.g 1120 Hamburg-Hasselbrook - Hamburg Hbf)

    # getter and setter methods
    @property
    def element_type(self) -> str:
        """Returns the type as string.

        Returns
        -------
        type: Layout
            The type as string."""
        return 'TrackSegment'

    @property
    def layout(self) -> Layout:
        """Computes the rail layout of track segment.

        Returns
        -------
        layout: Layout
            The rail layout of the track segment."""
        return self._layout

    @property
    def operational_point_from(self) -> OperationalPoint:
        """Returns the operational point at the start of track segment.

        Returns
        -------
        operational_point_from: OperationalPoint
            The operational point at the start of track segment."""
        return self._operational_point_from

    @property
    def operational_point_to(self) -> OperationalPoint:
        """Returns the operational point at the end of track segment.

        Returns
        -------
        operational_point_to: OperationalPoint
            The operational point at the end of track segment."""
        return self._operational_point_to

    @property
    def properties(self) -> Dict[str, str | float | int]:
        """Returns the properties dictionary. This dictionary contains all information for track segment (Richtungsgleis / eingleisig) from ISR.

        Returns
        -------
        properties: Dict[str, str | float | int]
            Dictionary with track segment information from ISR."""
        return self._properties

    @property
    def properties_2(self) -> Dict[str, str | float | int] | None:
        """Returns the properties dictionary. This dictionary contains all information for track segment (Gegengleis, if exists) from ISR.

        Returns
        -------
        properties: Dict[str, str | float | int] | None
            Dictionary with track segment information from ISR for Gegengleis. If track does not have a Gegengleis, None is returned.
        """
        return self._properties_2

    @properties.setter
    def properties(self, value: Dict[str, str | float | int]):
        """Sets properties dictionary for Richtungsgleis / eingleisig."""
        self._properties = value

    @properties_2.setter
    def properties_2(self, value: Dict[str, str | float | int]):
        """Sets properties dictionary for Gegengleis."""
        self._properties_2 = value

    @property
    def properties_info(self) -> Dict[str, Tuple[str, str]]:
        """Returns the properties info dictionary. This dictionary contains all information about attributes contained in 'properties'.

        Returns
        -------
        properties: Dict[str, str | float | int]
            Dictionary withinformation about attributes in 'properties'."""
        return self._properties_info

    @property
    def lines(self) -> list[LineString]:
        """Returns the lines of the track segment as list of LineString.

        Returns
        -------
        rails: list[LineString]
            The lines of the track segment.
        """
        return self._lines

    @property
    def lines_1(self):
        """Returns all rails of the track segment running in direction 1.

        Returns
        -------
        rails: list[LineString]
            All rails of the track segment running in direction 1.
        """
        layout = self.layout
        return [self.lines[i] for i in range(len(self.lines)) if i in layout['direction_1'][0]]

    @property
    def lines_2(self):
        """Returns all rails of the track segment running in direction 2.

        Returns
        -------
        rails: list[LineString]
            All rails of the track segment running in direction 2.
        """
        layout = self.layout
        return [self.lines[i] for i in range(len(self.lines)) if i in layout['direction_2'][0]]

    @property
    def rails(self) -> list[Rail]:
        """Returns the rails of the track segment as list of LineString.

        Returns
        -------
        rails: list[LineString]
            The rails of the track segment.
        """
        return self._rails

    @property
    def rails_1(self):
        """Returns all rails of the track segment running in direction 1.

        Returns
        -------
        rails: list[LineString]
            All rails of the track segment running in direction 1.
        """
        layout = self.layout
        return [self.rails[i] for i in range(len(self.rails)) if i in layout['direction_1'][0]]

    @property
    def rails_2(self):
        """Returns all rails of the track segment running in direction 2.

        Returns
        -------
        rails: list[LineString]
            All rails of the track segment running in direction 2.
        """
        layout = self.layout
        return [self.rails[i] for i in range(len(self.rails)) if i in layout['direction_2'][0]]

    @property
    def length_nominal(self) -> float:
        """The nominal length of track segment in kilometer according to official kilometrage, in km."""
        return self._length

    @property
    def length_nominal_isr(self) -> float:
        """The nominal length of track segment as stated in ISR, in km."""
        return self._properties['ALG_LAENGE_ABSCHNITT']

    @property
    def length_geometric(self) -> float:
        """The geometric length of track segment in kilometer according to actual geometry of track segment."""
        # TODO
        return self.kilometrage.length_nominal_km

    @property
    def length_geometric_nominal(self) -> float:
        """The nominal geometric length of track segment in kilometer according to Deutsche Bahn."""
        # TODO
        pass

    @property
    def length_nominal_m(self) -> float:
        """The nominal length of track segment in kilometer according to official kilometrage, in m."""
        return self._length * 1000

    @property
    def kilometrage(self) -> Kilometrage:
        """Returns the kilometrage instance."""
        return self._kilometrage

    @property
    def km_from_db(self) -> str:
        """Kilometrage value at operational_point_from, in DB format."""
        return str(self._km_from)

    @property
    def km_to_db(self) -> str:
        """Kilometrage value at operational_point_to, in DB format."""
        return str(self._km_to)

    @property
    def km_from(self) -> float:
        """Kilometrage value at operational_point_from, as float (float value includes overlength or underlength)."""
        return round(db_km_to_km(self.km_from_db), 3)

    @property
    def km_to(self) -> float:
        """Kilometrage value at operational_point_to, as float (float value includes overlength or underlength)."""
        return round(db_km_to_km(self.km_to_db), 3)

    @property
    def id(self) -> int:
        """ID of segment in ISR."""
        return int(self._id)

    @property
    def name(self) -> str:
        """Name of track segment (station_from - station_to)."""
        return str(self._name)

    @property
    def track_nr(self) -> int:
        """Track number the track segment belongs to."""
        return int(self._track_nr)

    @property
    def incline_dict(self) -> dict:
        """Returns dictionary with incline information."""
        return self._incline_dict

    @overload
    def km(self, point: Point, direction: Union[int, str], return_type: Literal['db']) -> str:
        ...

    @overload
    def km(self, point: Point, direction: Union[int, str], return_type: Literal['float']) -> float:
        ...

    def km(
        self, point: Point, direction: Union[int, str] = 0, return_type: Literal['db', 'float'] = 'float'
    ) -> Union[str, float]:
        """Compute the kilometrage for given coordinates 'point', direction 'direction' and in the desired format 'return_type'."""

        if direction in ['0', '1', '2']:
            direction = int(direction)
        elif direction in [0, 1, 2]:
            pass
        else:
            raise ValueError(f'Received direction={direction}, direction must be 0, 1 or 2.')

        # get attributes
        lines = self.rails

        # assert point is on rails
        distances_to_lines = [line.distance(point) for line in lines]
        min_distance_m = np.min(distances_to_lines)

        km = self.kilometrage.km(point=point, direction=direction, return_type=return_type)

        return km

    def point(self, km: float, direction: int) -> Point:
        """Compute the kilometrage for given coordinates 'point', direction 'direction' and in the desired format 'return_type'."""

        # assert km's are in range
        # assert (km >= self.km_from) and (km <= self.km_to), ValueError(f'Given kilometer must be in range of track segment: {self.km_from} - {self.km_to} km.')

        # find next smaller km value given 'km'
        km_values = self.kilometrage.values_in_direction(direction=direction)
        km_values = np.unique([round(db_km_to_km(km), 3) for km in km_values]).tolist()
        next_smaller_idx = bisect.bisect(km_values, km)

        # select corresponding kilometrage line
        idx_dir = 0 if direction < 2 else 1
        if idx_dir == len(self.kilometrage.lines):
            idx_dir = 0
        if km == km_values[-1] or (len(km_values) > 1 and km_values[-1] > self.km_to and km == km_values[-2]):
            next_smaller_idx -= 1
        if km < self.km_from:
            next_smaller_idx = 1

        if len(self.kilometrage.lines[idx_dir]) == 1:
            km_line = self.kilometrage.lines[idx_dir][0]
        else:
            if next_smaller_idx - 1 == len(self.kilometrage.lines[idx_dir]):
                km_line = self.kilometrage.lines[idx_dir][next_smaller_idx - 2]
            else:
                km_line = self.kilometrage.lines[idx_dir][next_smaller_idx - 1]

        # get diff: km_diff = km - smaller_km
        km_diff = km - km_values[next_smaller_idx - 1]
        if km_diff < 0:
            km_diff = 0

        # project
        p = km_line.interpolate(km_diff * 1000)

        # test_km = self.km(p, direction=direction, return_type='float')

        return p

    def distance(self, geom: T) -> float:
        """Returns the shortest distance of track segment to another shapely object 'geom'.

        Returns
        -------
        min_distance_m: float
            The shortest distance between 'self' and 'geom', in m.
        """
        rails = self.rails
        distances_m = []
        for rail in rails:
            distance_m = rail.distance(geom)
            distances_m.append(distance_m)
        min_distance_m = np.min(distances_m)
        return min_distance_m

    def _compute_kilometrage(self):
        """Computes kilometrage lines for track segment and uses them to instantiate instance of class Kilometrage, which is set as attribute and handles everything kilometrage related.

        Kilometrage lines are computed either as middle line of rail (monorails) or as middle line of two parallel rails (parallel rails). This is done for each segment of track segment seperately, afterwards the kilometrage lines are joined to one line.
        """

        # fetch attributes
        lines = self._lines
        layout = self._layout
        n_km_lines = self._n_km_lines

        if n_km_lines != 1 and n_km_lines != 2:
            raise ValueError(f'invalid value for n_km_lines: {n_km_lines} (must be 1 or 2)')

        # isr preprocessor
        isr_preprocessor = ISRPreprocessor(
            operational_point_from=self.operational_point_from, operational_point_to=self.operational_point_to
        )

        # all paths in direction 1
        layout_rails = layout['direction_1']
        n_paths = len(layout_rails)
        # all parallel rails
        index_parallel = layout['parallel']
        # all monorails
        index_monorail = layout['monorails']

        # path with least number of rails (= path without sidings)
        if n_paths >= 1:
            path = min(layout_rails, key=len)
        else:
            raise ValueError(f'Layout contains no path. Computing kilometrage line not possible.')

        if n_km_lines == 1:
            # if path is a single monorail
            if (len(path) == 1) and (path[0] in index_monorail):
                # take this rail as kilometrage line
                kilometrage_line = [lines[path[0]]]

            else:
                # for monorail: append as is
                # for parallel rails: compute middle line

                kilometrage_lines: list[LineString] = []
                for indice in path:
                    # parallel lines
                    if indice in flatten_list(index_parallel):
                        for index in index_parallel:
                            if indice in index:
                                # get parallel lines
                                line_1 = lines[index[0]]
                                line_2 = lines[index[1]]

                                difference_m = line_1.length - line_2.length
                                d_start_m = Point(line_1.coords[0]).distance(Point(line_2.coords[0]))
                                d_end_m = Point(line_1.coords[-1]).distance(Point(line_2.coords[-1]))

                                one_line_exists_with_length_close_to_km = (
                                    abs(self.length_nominal_m - line_1.length) < 3
                                    and abs(self.length_nominal_m - line_2.length) > 3
                                ) or (
                                    abs(self.length_nominal_m - line_2.length) < 3
                                    and abs(self.length_nominal_m - line_1.length) > 3
                                )  # for two lines only
                                if abs(difference_m) > 15 and (
                                    (d_start_m > 10)
                                    or (d_end_m) > 10
                                    or (len(lines) == 2 and one_line_exists_with_length_close_to_km)
                                ):
                                    # identify longer and shorter line
                                    if difference_m > 0:
                                        line_l = line_1
                                        line_s = line_2
                                    else:
                                        line_l = line_2
                                        line_s = line_1

                                    # if parallel lines have different lengths, take that line as km reference that results in less kilometrage error
                                    # testing is done by summing over all line lengths in path
                                    path_length_diff_longer_line = abs(
                                        self.length_nominal_m
                                        - line_l.length
                                        - sum(
                                            [lines[k].length for k in path if lines[k] != line_l and lines[k] != line_s]
                                        )
                                    )
                                    path_length_diff_shorter_line = abs(
                                        self.length_nominal_m
                                        - line_s.length
                                        - sum(
                                            [lines[k].length for k in path if lines[k] != line_s and lines[k] != line_l]
                                        )
                                    )

                                    if path_length_diff_longer_line <= path_length_diff_shorter_line:
                                        kilometrage_lines.append(line_l)
                                    else:
                                        kilometrage_lines.append(line_s)

                                else:
                                    # compute middle line
                                    middle_line = self._compute_middle_line(line_1=line_1, line_2=line_2)
                                    km_coords = middle_line.coords[:]
                                    angles = [
                                        angle_between(
                                            np.array(km_coords[i + 1]) - np.array(km_coords[i]),
                                            np.array(km_coords[i + 2]) - np.array(km_coords[i + 1]),
                                        )
                                        for i in range(len(km_coords) - 2)
                                    ]

                                    if len(angles) > 0:
                                        if max(angles) > 30 and len(index_parallel) == 1:
                                            offical_length_m = self.length_nominal_m
                                            diff1 = abs(offical_length_m - line_1.length)
                                            diff2 = abs(offical_length_m - line_2.length)
                                            min_diff = np.argmin([diff1, diff2])
                                            middle_line = [line_1, line_2][min_diff]

                                    # append
                                    kilometrage_lines.append(middle_line)

                                # special case: two monorails after another that are not connected (can be partly parralel, with one line continuing further than the other)
                                if (
                                    (path == [0, 1] or path == [1, 0])
                                    and lines[path[0]].length < self.length_nominal_m
                                    and lines[0].length + lines[1].length >= self.length_nominal_m
                                    and len(lines) == 2
                                ):
                                    # lines in path direction
                                    line_1 = lines[path[0]]
                                    line_2 = lines[path[1]]

                                    ep_projected_on_line_2 = nearest_points(Point(line_1.coords[-1]), line_2)[1]
                                    distance_along_line_2 = line_2.project(ep_projected_on_line_2, normalized=True)

                                    if distance_along_line_2 <= 0.9:
                                        line_2_segment = split(line_2, ep_projected_on_line_2)[1]
                                        kilometrage_lines.append(line_2_segment)

                    # monorails
                    else:
                        # append
                        kilometrage_lines.append(lines[indice])

                # kilometrage line: merge segments and smoothen links
                kilometrage_line = isr_preprocessor.merge_lines_and_smoothen_links(
                    lines=kilometrage_lines, merge_km_line=True
                )

            self._kilometrage = Kilometrage(km_lines=[kilometrage_line], enhance_kilometrage=self._enhance_kilometrage)

        if n_km_lines == 2:
            # merge segments in each direction to create two km lines
            layout_rails_1 = layout['direction_1']
            # path with least number of rails (= path without sidings)
            if len(layout_rails_1) >= 1:
                path1 = min(layout_rails_1, key=len)
            else:
                raise ValueError('Should not happen')  # TODO

            layout_rails_2 = layout['direction_2']
            if len(layout_rails_2) >= 1:
                path2 = min(layout_rails_2, key=len)
            else:
                raise ValueError('Should not happen')  # TODO
            if self.name in 'Mannheim Rbf Gr G, W 105 - Mannheim Rbf Gr K Kn, W 650':
                path1.reverse()
                path2.reverse()
            kilometrage_lines_1 = isr_preprocessor.merge_lines_and_smoothen_links(lines=lines, merge_index=[path1])
            kilometrage_lines_2 = isr_preprocessor.merge_lines_and_smoothen_links(
                lines=lines, merge_index=[path2[::-1]]
            )  # reverse due to sorting of lines

            self._kilometrage = Kilometrage(
                km_lines=[kilometrage_lines_1, kilometrage_lines_2], enhance_kilometrage=self._enhance_kilometrage
            )

    def incline(self, km: float, direction: int):
        """
        Returns incline of track segment at postion given by 'km'. Inclines are taken from ISR and are assumed piecewise constant, as stated in ISR.

        TODO
        """

        # if incline information is available for track segment
        if not isinstance(self.incline_dict['inclines_dir1'], type(None)):
            km = round(km, 3)

            # assert km is in range of track
            # assert km >= self.km_from, ValueError(f"'km' must be in range of track segment: {self.km_from} - {self.km_to}") # km <= self.km_to and

            # assert direction
            assert direction in [0, 1, 2], ValueError('Direction must be 0, 1 or 2.')

            # find next smaller km to 'km' in incline profile (index)
            if direction == 0 or direction == 1 or isinstance(self.incline_dict['inclines_dir2'], type(None)):
                # get incline value by index
                next_smaller_idx = bisect.bisect(self.incline_dict['incline_kms_dir1'], km)
                value = self.incline_dict['inclines_dir1'][next_smaller_idx - 1]
            else:
                # get incline value by index
                next_smaller_idx = bisect.bisect(self.incline_dict['incline_kms_dir2'], km)
                value = self.incline_dict['inclines_dir2'][next_smaller_idx - 1]

            if direction == 2:
                value = -value

        # no incline information available
        else:
            value = np.nan

        return value

    def incline_profile(self, km_from: float, km_to: float):
        """Returns the incline profile of track segment, starting from 'km_from' and ending at 'km_to'."""

        km_from = round(km_from, 3)
        km_to = round(km_to, 3)

        # assert km is in range of track
        #        assert km_from >= self.km_from, ValueError(f"'km' must be in range of track segment: {self.km_from} - {self.km_to}") # km_to <= self.km_to and

        # moving direction
        km_diff = km_to - km_from
        if km_diff > 0:
            direction = 1
        elif km_diff < 0:
            direction = 2
            km_from, km_to = km_to, km_from
        else:
            raise ValueError('Values for km_from and km_to are equal.')

        if direction == 1:
            # if incline information is available for track segment
            if not isinstance(self.incline_dict['inclines_dir1'], type(None)):
                # find next smaller km to 'km' in incline profile (index)

                # get incline profile for direction 1
                kms_dir1 = self.incline_dict['incline_kms_dir1']
                inclines_dir1 = self.incline_dict['inclines_dir1']
                if km_from < kms_dir1[0]:
                    next_smaller_idx = 1
                else:
                    next_smaller_idx = bisect.bisect(kms_dir1, km_from)
                next_higher_idx = bisect.bisect(kms_dir1, km_to)

                kms = [km_from] + kms_dir1[next_smaller_idx:next_higher_idx] + [km_to]
                inclines = inclines_dir1[next_smaller_idx - 1 : next_higher_idx]
                inclines += [inclines[-1]]

            elif not isinstance(self.incline_dict['inclines_dir2'], type(None)):
                # get incline profile for direction 1 if direction 2 is not available (-> "Auf Anfrage")
                kms_dir1 = self.incline_dict['incline_kms_dir2']
                inclines_dir1 = self.incline_dict['inclines_dir2']
                if km_from < kms_dir1[0]:
                    next_smaller_idx = 1
                else:
                    next_smaller_idx = bisect.bisect(kms_dir1, km_from)
                next_higher_idx = bisect.bisect(kms_dir1, km_to)

                kms = [km_from] + kms_dir1[next_smaller_idx:next_higher_idx] + [km_to]
                inclines = inclines_dir1[next_smaller_idx - 1 : next_higher_idx]
                inclines += [inclines[-1]]

            # no incline information available
            else:
                kms, inclines = [km_from, km_to], [np.nan, np.nan]

        else:  # direction = 2
            # if incline information is available for track segment
            if not isinstance(self.incline_dict['inclines_dir2'], type(None)):
                # get incline profile for direction 2 (negate and reverse)
                kms_dir2 = self.incline_dict['incline_kms_dir2']
                inclines_dir2 = self.incline_dict['inclines_dir2']
                if km_from < kms_dir2[0]:
                    next_smaller_idx = 1
                else:
                    next_smaller_idx = bisect.bisect(kms_dir2, km_from)
                next_higher_idx = bisect.bisect(kms_dir2, km_to)

                kms = [km_from] + kms_dir2[next_smaller_idx:next_higher_idx] + [km_to]
                inclines = [
                    -inc for inc in inclines_dir2[next_smaller_idx - 1 : next_higher_idx]
                ]  # negate to invert incline for opposite direction
                kms = kms[::-1]
                inclines = inclines[::-1]
                inclines += [inclines[-1]]

            elif not isinstance(self.incline_dict['inclines_dir1'], type(None)):
                # get incline profile for direction 1 if direction 2 is not available (-> "Auf Anfrage")
                kms_dir2 = self.incline_dict['incline_kms_dir1']
                inclines_dir2 = self.incline_dict['inclines_dir1']
                if km_from < kms_dir2[0]:
                    next_smaller_idx = 1
                else:
                    next_smaller_idx = bisect.bisect(kms_dir2, km_from)
                next_higher_idx = bisect.bisect(kms_dir2, km_to)

                kms = [km_from] + kms_dir2[next_smaller_idx:next_higher_idx] + [km_to]
                inclines = [
                    -inc for inc in inclines_dir2[next_smaller_idx - 1 : next_higher_idx]
                ]  # negate to invert incline for opposite direction
                kms = kms[::-1]
                inclines = inclines[::-1]
                inclines += [inclines[-1]]

            # no incline information available
            else:
                kms, inclines = [km_to, km_from], [np.nan, np.nan]

        return kms, inclines

    def compute_incline(self) -> dict:
        """
        Computes incline profile of track segment.

        TODO
        """

        # main direction
        if not (self.properties['INF_STEIGUNGSPROFIL'] == 'auf Anfrage'):
            if not ('(' in self.properties['INF_STEIGUNGSPROFIL']):  # single incline
                kms_incline_dir1 = [self.km_from]
                inclines_dir1 = [float(self.properties['INF_STEIGUNGSPROFIL'])]
            else:
                # extract kilometrage and incline values from track segment
                kms_incline_dir1 = [
                    float(string.split('(')[1].split(')')[0])
                    for string in self.properties['INF_STEIGUNGSPROFIL'].strip().split(';')
                ]
                inclines_dir1 = [
                    float(string.split('(')[0].strip())
                    for string in self.properties['INF_STEIGUNGSPROFIL'].strip().split(';')
                ]

        else:
            kms_incline_dir1 = None
            inclines_dir1 = None

        # if second direction exists
        if self.properties_2 and not (self.properties_2['INF_STEIGUNGSPROFIL'] == 'auf Anfrage'):
            if not ('(' in self.properties_2['INF_STEIGUNGSPROFIL']):  # single incline
                kms_incline_dir2 = [self.km_from]
                inclines_dir2 = [float(self.properties_2['INF_STEIGUNGSPROFIL'])]
            else:
                kms_incline_dir2 = [
                    float(string.split('(')[1].split(')')[0])
                    for string in self.properties_2['INF_STEIGUNGSPROFIL'].strip().split(';')
                ]
                inclines_dir2 = [
                    float(string.split('(')[0].strip())
                    for string in self.properties_2['INF_STEIGUNGSPROFIL'].strip().split(';')
                ]
        else:
            kms_incline_dir2 = None
            inclines_dir2 = None

        incline_dict = {
            'incline_kms_dir1': kms_incline_dir1,
            'inclines_dir1': inclines_dir1,
            'incline_kms_dir2': kms_incline_dir2,
            'inclines_dir2': inclines_dir2,
        }

        return incline_dict

    def _compute_middle_line(self, line_1: LineString, line_2: LineString) -> LineString:
        """Returns the middle line between two input LineStrings.

        Computes the middle line by finding orthogonal projections of start and end points,
        and interpolating points along the segments. Result is a LineString.

        Parameters
        ----------
        line_1 : LineString
            First input LineString.
        line_2 : LineString
            Second input LineString.

        Returns
        -------
        middle_line : LineString
            The middle line between 'line_1' and 'line_2'.
        """

        # compute mean of start points while considering orthogonality
        point1 = Point(line_1.coords[0])
        point2 = Point(line_2.coords[0])

        point12 = nearest_points(point1, line_2)[1]
        point21 = nearest_points(point2, line_1)[1]

        if point12 != point2:
            line = LineString([point1, point12])
        elif point21 != point1:
            line = LineString([point21, point2])
        else:
            line = LineString([point1, point2])
        start_point = line.interpolate(distance=0.5, normalized=True)

        # compute mean of end points
        point1 = Point(line_1.coords[-1])
        point2 = Point(line_2.coords[-1])
        point12 = nearest_points(point1, line_2)[1]
        point21 = nearest_points(point2, line_1)[1]

        # if orthagonal projection of line1's end point is not line2's end point (not orthagonal)
        if point12 != point2:
            # orthagonal line w.r.t. rail(s)
            line = LineString([point1, point12])

        # if orthagonal projection of line2's end point is not line1's end point (not orthagonal)
        elif point21 != point1:
            # orthagonal line w.r.t. rail(s)
            line = LineString([point21, point2])

        # else end points are on the orthagonal w.r.t rails (orthagonal)
        else:
            # orthagonal line w.r.t. rail(s)
            line = LineString([point1, point2])

        end_point = line.interpolate(distance=0.5, normalized=True)

        # initalize new coordinates with start point
        new_line_coords: list[Tuple[float, float]] = [start_point]

        # sampling points from first line
        line_1_coords = [Point(coord) for coord in line_1.coords[1:-1]]

        # nearest points on other line (done in segments avoid problems with circular rails)
        n_coords = len(line_1_coords)
        n = n_coords // 10
        all_points_line_1 = []
        line_2_nearest = []

        for i in range(0, n + 1):
            points_line_1 = [Point(coord) for coord in line_1_coords[i * 10 : (i + 1) * 10]]

            if n_coords - len(line_2.coords) > 50:
                line_2_segment = line_2
            else:
                line_2_segment = LineString(line_2.coords[max(i - 4, 0) * 10 : (i + 5) * 10])

            line_2_nearest_ = nearest_points(points_line_1, line_2_segment)[1]

            all_points_line_1 += points_line_1
            line_2_nearest += list(line_2_nearest_)
            assert len(line_2_segment.coords[:]) > 1, f'wrong window size for line segments (i = {i})'

        for point_1, point_2 in zip(all_points_line_1, line_2_nearest):
            # create linestring
            line_string = LineString([point_1, point_2])
            # get middle point
            middle_point = line_string.interpolate(distance=0.5, normalized=True)
            new_line_coords.append(middle_point)

        # add end point and create line string
        new_line_coords.append(end_point)
        middle_line = LineString(new_line_coords)

        return middle_line

    def _compute_track_segment_layout(self, lines: list[LineString]) -> Layout:
        """Computes the layout of a track segment based on input LineStrings.

        Adjusts for specific cases, calculates rail type indexes, and organizes rails
        into layout dictionaries for both directions.

        Parameters
        ----------
        lines : list[LineString]
            List of LineStrings representing track segments.

        Returns
        -------
        layout : Layout
            Dictionary containing layout information for the track segment.
        """

        # special case: wrong direction due to circular track
        if self.name == 'StrUeb_4410_4411 - StrUeb_4411_4413' and self.track_nr == 4411:
            self._lines = [line.reverse() for line in lines][::-1]
        # rail segment missing in ISR (6083: StrUeb_6081_6083 - Berlin-Karow West)
        if 'StrUeb_6081_6083' in self.name and self.track_nr == 6083:
            lines[0] = LineString([self.operational_point_from] + list(lines[0].coords))

        # compute indexes
        (
            index_parallel,
            index_monorail,
            index_siding,
            index_direction_1,
            index_direction_2,
        ) = self._compute_rail_type_indexes(lines=lines)

        # get rail layout in both directions
        layout_direction_1 = []
        layout_direction_2 = []
        for i in range(0, len(lines)):
            # monorails or sidings
            if not i in flatten_list(index_parallel):
                # sidings (Abzweigendes Gleis)
                if i in flatten_list(index_siding)[1::2]:
                    siding_array = np.array(index_siding)
                    idx = np.where(siding_array == i)[0][0]
                    layout_direction_1.append(index_siding[idx])
                    layout_direction_2.append(index_siding[idx][::-1])

                # monorails (eingleisig)
                else:
                    layout_direction_1.append(i)
                    layout_direction_2.append(i)

            # parallel rails
            else:
                # direction track (Richtungsgleis)
                if i in index_direction_1:
                    layout_direction_1.append(i)

                # opposite track (Gegengleis)
                elif i in index_direction_2:
                    layout_direction_2.append(i)

        # reverse layout for opposite track
        layout_direction_2 = layout_direction_2[::-1]

        # format lists
        layout_contains_list = any([True for val in layout_direction_1 if type(val) == list])
        if layout_contains_list:
            layout_direction_1_new = [[], []]
            for val in layout_direction_1:
                if type(val) != list:
                    if not (val in layout_direction_1_new[0]):
                        layout_direction_1_new[0].append(val)
                    if not (val in layout_direction_1_new[1]):
                        layout_direction_1_new[1].append(val)
                else:
                    if not (val[0] in layout_direction_1_new[0]):
                        layout_direction_1_new[0].append(val[0])
                    if not (val[1] in layout_direction_1_new[1]):
                        layout_direction_1_new[1].append(val[1])

        else:
            layout_direction_1_new = [layout_direction_1]

        layout_contains_list = any([True for val in layout_direction_2 if type(val) == list])

        if layout_contains_list:
            layout_direction_2_new = [[], []]

            for val in layout_direction_2:
                if type(val) != list:
                    if not (val in layout_direction_2_new[0]):
                        layout_direction_2_new[0].append(val)
                    if not (val in layout_direction_2_new[1]):
                        layout_direction_2_new[1].append(val)
                else:
                    if not (val[0] in layout_direction_2_new[0]):
                        layout_direction_2_new[0].append(val[0])
                    if not (val[1] in layout_direction_2_new[1]):
                        layout_direction_2_new[1].append(val[1])
                layout_direction_2_new.sort(key=len)
        else:
            layout_direction_2_new = [layout_direction_2]

        # another special case: #
        if self.name == 'Seelze Mitte - Ahlem' and self.track_nr == 1750:
            layout_direction_1_new = [[0, 3]]
            layout_direction_2_new = [[2, 1]]
            index_direction_1[0] = 0
            index_direction_2[0] = 1
            index_parallel[0] = (0, 1)

        # layout dict
        layout: Layout = {
            'direction_1': layout_direction_1_new,  # all paths in direction 1
            'direction_2': layout_direction_2_new,  # all paths in direction 2
            'monorails': index_monorail,  # all monorails (bidrectional)
            'direction_1_rails': index_direction_1,  # all rails with direction 1 (unidirectional)
            'direction_2_rails': index_direction_2,  # all rails with direction 2 (unidirectional)
            'parallel': index_parallel,  # all rails which are parallel
            'sidings': index_siding,  #
        }

        return layout

    def _compute_rail_type_indexes(
        self, lines: list[LineString]
    ) -> Tuple[list[Tuple[int, int]], list[int], list[Tuple[int, int]], list[int], list[int]]:
        """Computes indexes for different rail types within a track segment.

        Analyzes the relationships between lines to determine parallel tracks, sidings,
        and the direction of each track.

        Parameters
        ----------
        lines : list[LineString]
            List of LineStrings representing track segments.

        Returns
        -------
        indexes : Tuple[list[Tuple[int, int]], list[int], list[Tuple[int, int]], list[int], list[int]]
            Tuple containing indexes for parallel lines, monorails, sidings, and
            direction-specific tracks.
        """

        index_parallel = []  # index of parallel lines
        index_siding = []  # index of sidings
        index_direction_1 = []  # index of rails operating in direction 1
        index_direction_2 = []  # index of rails operating in direction 2

        if len(lines) == 0:
            raise ValueError(f'Input list "lines" is empty.')

        elif len(lines) == 1:
            index_monorail = [0]

        else:
            # iterate all combinations of lines
            for i in range(0, len(lines)):
                for j in range(0, len(lines)):
                    # only one combination of each two lines, and no same lines
                    if j > i:
                        # start and end points of lines i,j
                        i_start_point = Point(lines[i].coords[0])
                        j_start_point = Point(lines[j].coords[0])
                        i_end_point = Point(lines[i].coords[-1])
                        j_end_point = Point(lines[j].coords[-1])

                        # distance between start points and end points
                        distance_ij_start = i_start_point.distance(j_start_point)
                        distance_ij_end = i_end_point.distance(j_end_point)
                        if TYPE_CHECKING:  # for type hint only
                            distance_ij_start = float(distance_ij_start)
                            distance_ij_end = float(distance_ij_end)

                        # condition for sidings (assuming parallel)
                        # if lines are parallel, but have but touch at exactly one point: siding
                        d_start_m = Point(lines[i].coords[0]).distance(Point(lines[j].coords[0]))
                        d_end_m = Point(lines[i].coords[-1]).distance(Point(lines[j].coords[-1]))
                        siding_condition = lines[i].touches(lines[j]) and ((d_start_m > 10) and (d_end_m) > 10)

                        # parallel lines
                        if lines_are_parallel(line1=lines[i], line2=lines[j]):
                            # parallel tracks

                            # line string between start points
                            line_string_1 = LineString([i_start_point, j_start_point])
                            line_string_2 = LineString([j_end_point, i_end_point])

                            # extend line string by factor 1.5
                            line_string_extended_1 = affinity.scale(
                                geom=line_string_1, xfact=1.5, yfact=1.5, origin=i_start_point
                            )
                            line_string_extended_2 = affinity.scale(
                                geom=line_string_2, xfact=1.5, yfact=1.5, origin=j_end_point
                            )

                            # rotate line string clockwise by 10 degrees
                            line_string_rotated_1 = affinity.rotate(
                                geom=line_string_extended_1, angle=-10, origin=i_start_point
                            )
                            line_string_rotated_2 = affinity.rotate(
                                geom=line_string_extended_2, angle=-10, origin=j_end_point
                            )

                            line_i_direction = np.array(lines[i].coords[1]) - np.array(lines[i].coords[0])
                            line_1_direction = np.array(line_string_1.coords[1]) - np.array(line_string_1.coords[0])
                            line_j_direction = np.array(lines[j].coords[-1]) - np.array(lines[j].coords[-2])
                            line_2_direction = np.array(line_string_2.coords[1]) - np.array(line_string_2.coords[0])
                            angle_ij = angle_between(line_i_direction, line_1_direction)
                            angle_ji = angle_between(line_j_direction, line_2_direction)

                            special_1 = (
                                self.track_nr == 6145 and self.name == 'Grünauer Kreuz West - Grünauer Kreuz Süd'
                            )
                            special_2 = (
                                self.track_nr == 6072 and self.name == 'Biesdorfer Kreuz West - StrUeb_6070_6072'
                            )
                            special_case = True if special_1 or special_2 else False

                            # sidings
                            if siding_condition:
                                # line lengths
                                length_i = lines[i].length
                                length_j = lines[j].length
                                # shorter line is treated as siding of longer line
                                # longer line is represented by first indice in list
                                # shorter line is represented by second indice in list
                                if length_i >= length_j:
                                    index_siding.append([i, j])
                                else:
                                    index_siding.append([j, i])

                            elif (
                                (line_string_rotated_1.intersects(lines[j]) and abs(angle_ij - 90) < 20)
                                or (line_string_rotated_2.intersects(lines[i]) and abs(angle_ji - 90) < 20)
                                or (
                                    (np.isnan(angle_ji) and line_string_rotated_1.intersects(lines[j]))
                                    or (np.isnan(angle_ij) and line_string_rotated_2.intersects(lines[i]))
                                )
                                or special_case == True
                            ):
                                # i: direction track
                                # j: opposite track
                                index_direction_1.append(i)
                                index_direction_2.append(j)
                                index_parallel.append((i, j))

                            else:
                                # i: opposite track
                                # j: direction track
                                index_direction_2.append(i)
                                index_direction_1.append(j)
                                index_parallel.append((j, i))

            # index of monorails
            index_monorail = [i for i in range(0, len(lines)) if not i in flatten_list(index_parallel)]

        return index_parallel, index_monorail, index_siding, index_direction_1, index_direction_2

    def _find_index_of_parallel_rail(self, rail_indice: int, index_parallel: list[Tuple[int, int]]) -> int:
        """Finds the index of the parallel rail for a given rail index.

        Parameters
        ----------
        rail_indice : int
            Index of the rail for which the parallel rail index is sought.
        index_parallel : list[Tuple[int, int]]
            List of tuples representing indices of parallel rails.

        Returns
        -------
        parallel_index : int
            Index of the parallel rail for the given 'rail_indice'.
        """

        if rail_indice in flatten_list(index_parallel):
            for group in index_parallel:
                if rail_indice in group:
                    if rail_indice == group[0]:
                        i = group[1]
                    else:
                        i = group[0]
                    break
        else:
            raise ValueError(f"Rail for given indice 'rail_indice' = {rail_indice} does not have a parallel rail.")

        return i

    @classmethod
    def from_track_and_name(
        cls, track: int, name: str, to_km: float, allow_previous: bool = True, enhance_kilometrage: bool = True
    ) -> Union[TrackSegment, None]:
        """Creates a TrackSegment instance from track number and name.

        Attempts to load the TrackSegment from cache; if not found, queries the parameters and creates a new instance.

        Parameters
        ----------
        track : int
            Track number.
        name : str
            Name of the track segment.
        allow_previous : bool, optional
            Flag to allow loading from previous cache, by default True.

        Returns
        -------
        track_segment : Union[TrackSegment, None]
            The TrackSegment instance if found, else None.
        """

        # paths
        project_path = Path(__file__).parent.parent.parent.parent

        if any([ex_name in name for ex_name in ISR_EXCEPTIONAL_STATION_NAMES]):
            name = extract_station_names(track_name_string=name, is_op_name=True).replace('**', '*')
            if name[-1] == '*':
                name = name[:-1]
            # name = f'{from_name} - {to_name}'
        km_type = 'raw' if enhance_kilometrage == False else 'enh'
        if int(track) < 4930:
            file_path = project_path / f'cache/track_segments/{track}_{name}_{km_type}.pickle'
        else:
            file_path = project_path / f'cache/track_segments_continued/{track}_{name}_{km_type}.pickle'

        matching_files = sorted(file_path.parent.glob(f'{track}_{name}_{km_type}.pickle'))

        if len(matching_files) == 1:
            file_path = matching_files[0]
            # load from file
            with open(file_path, "rb") as file:
                track_segment: TrackSegment = pickle.load(file)

            return track_segment

        else:
            # query and create instance
            parameter_tuple = query_track_segment(
                args=[track, name], to_km=to_km, allow_previous=allow_previous, enhance_kilometrage=enhance_kilometrage
            )

            if parameter_tuple:
                return cls(*parameter_tuple)
            else:
                return None

    def pickle_self_to_cache(self):
        """Pickles self to cache."""
        if self.track_nr > 4929:
            self.cache_path = Path(self.cache_path.as_posix().replace('track_segments', 'track_segments_continued'))
        with open(self.cache_path, "wb") as file:
            pickle.dump(obj=self, file=file, protocol=pickle.HIGHEST_PROTOCOL)

    def _integrate_kilometrage_points(self):
        """
        Integrate kilometrage points and adjust geometric lengths for the track segment.

        This method performs the integration of kilometrage points obtained from 'db-geo-streckennetz'
        for the given track segment. It also adjusts the geometric lengths based on the integrated
        kilometrage points and updates the kilometrage information.

        Notes
        -----
        The method processes kilometrage points, adjusts geometric lengths, and integrates
        information into the kilometrage. It handles various edge cases, including merging
        points with the kilometrage line, eliminating double entries, and correcting nominal
        geometric lengths for split segments. There are serveral problems when combining
        geometric information from ISR with kilometrage information from 'db-geo-streckennetz',
        which this functions tries to handle.
        """

        # get values
        track_km_dict = deepcopy(self.track_km_dict)
        track_length_dict = deepcopy(self.track_length_dict)
        segment_ids = track_km_dict.keys()
        op_from = self.operational_point_from
        op_to = self.operational_point_to
        km_from_db = self.km_from_db
        km_to_db = self.km_to_db

        # special cases
        if self.track_nr == 2670 and self.name == 'Köln-Mülheim Berliner Straße - Köln-Stammheim':
            km_to_db = '7,6 + 1487'  # stated as 7.4 + 1687, but points 7.5 and 7.6 exist before (km values dont align)
        elif self.track_nr == 2670 and self.name == 'Köln-Stammheim - Leverkusen Chempark Hp':
            km_from_db = '7,6 + 1487'  # consistency with next segment
        nominal_segment_length = (self.km_to - self.km_from) * 1000

        # set ids for direction cases
        if (f'{self.id}-0' in segment_ids) and self._enhance_kilometrage == True:
            segment_id_0 = f'{self.id}-0'
            segment_id_1 = None
            segment_id_2 = None
            self.has_additional_km_info = True
        elif (f'{self.id}-1') in segment_ids and self._enhance_kilometrage == True:
            segment_id_0 = None
            segment_id_1 = f'{self.id}-1'
            segment_id_2 = f'{self.id}-2'
            if not (f'{self.id}-2' in segment_ids):
                track_km_dict[segment_id_2] = track_km_dict[segment_id_1]
                track_length_dict[segment_id_2] = track_length_dict[segment_id_1]
            self.has_additional_km_info = True
        else:
            # no kilometrage info from dataset geo-streckennetz available for track segment
            logger.debug('No additional kilometrage information for track segment available')
            self.has_additional_km_info = False
            if self._n_km_lines == 1:
                self.kilometrage.add_km_point(
                    point=op_from, value=km_from_db, length=nominal_segment_length, direction=0
                )
                self.kilometrage.add_km_point(point=op_to, value=km_to_db, length=0, direction=0)
                self.kilometrage._km_lengths = [[nominal_segment_length]]
                self.kilometrage._km_lengths_type = [['computed']]
                self.kilometrage._has_km_info = False
            else:
                self.kilometrage.add_km_point(
                    point=op_from, value=km_from_db, length=nominal_segment_length, direction=1
                )
                self.kilometrage.add_km_point(
                    point=op_from, value=km_from_db, length=nominal_segment_length, direction=2
                )
                self.kilometrage.add_km_point(point=op_to, value=km_to_db, length=0, direction=1)
                self.kilometrage.add_km_point(point=op_to, value=km_to_db, length=0, direction=2)
                self.kilometrage._km_lengths = [[nominal_segment_length], [nominal_segment_length]]
                self.kilometrage._km_lengths_type = [['computed'], ['computed']]
                self.kilometrage._has_km_info = False
            return

        # only keep information for current track segment
        segment_id_list = [segment_id_0, segment_id_1, segment_id_2]
        track_km_dict = {key: track_km_dict[key] for key in segment_id_list if not isinstance(key, type(None))}
        track_length_dict = {key: track_length_dict[key] for key in segment_id_list if not isinstance(key, type(None))}

        # convert coordinates to points
        for direction, segment_id in enumerate(segment_id_list):
            if not segment_id:
                continue

            for i in range(len(track_km_dict[segment_id])):
                # iterate points for km and convert to Point
                point_list: list[Point] = []
                for j in range(len(track_km_dict[segment_id][i][1])):
                    if type(track_km_dict[segment_id][i][1][j]) != Point:
                        p = Point(json.loads(s=track_km_dict[segment_id][i][1][j])['coordinates'])
                    else:
                        p = track_km_dict[segment_id][i][1][j]
                    point_list.append(p)

                # replace string point list by list of Point objects
                track_km_dict[segment_id][i][1] = point_list

            # and add operational points
            if self._n_km_lines == 1:
                # direction 0
                distances = [op_from.distance(rail) for rail in self.rails]
                min_index = np.argmin(distances)
                rail_from = self.rails[min_index]

                distances = [op_to.distance(rail) for rail in self.rails]
                min_index = np.argmin(distances)
                rail_to = self.rails[min_index]

            else:
                if int(segment_id[-1]) == 1:
                    # direction 1
                    distances = [op_from.distance(rail) for rail in self.rails_1]
                    min_index = np.argmin(distances)
                    rail_from = self.rails_1[min_index]

                    distances = [op_to.distance(rail) for rail in self.rails_1]
                    min_index = np.argmin(distances)
                    rail_to = self.rails_1[min_index]

                else:
                    # direction 2
                    distances = [op_from.distance(rail) for rail in self.rails_2]
                    min_index = np.argmin(distances)
                    rail_from = self.rails_2[min_index]

                    distances = [op_to.distance(rail) for rail in self.rails_2]
                    min_index = np.argmin(distances)
                    rail_to = self.rails_2[min_index]

            # ops projected on rail
            if op_from.distance(rail_from) > 10:
                rail_from_ext = affinity.scale(
                    geom=LineString(rail_from.coords[:2]), xfact=5, yfact=5, origin=rail_from.coords[1]
                )
            else:
                rail_from_ext = rail_from

            if op_to.distance(rail_to) > 10:
                rail_to_ext = affinity.scale(
                    geom=LineString(rail_to.coords[-2:]), xfact=5, yfact=5, origin=rail_to.coords[-2]
                )
            else:
                rail_to_ext = rail_to

            op_from_on_rail = nearest_points(op_from, rail_from_ext)[1]
            op_to_on_rail = nearest_points(op_to, rail_to_ext)[1]

            # set value lists
            if op_from_on_rail.distance(rail_from) > 20:
                merge_from = 'before'
            else:
                merge_from = None
            if op_to_on_rail.distance(rail_to) > 20:
                merge_after = 'after'
            else:
                merge_after = None
            if op_from_on_rail.distance(op_from) > 15:
                op_from_ = op_from
            else:
                op_from_ = op_from_on_rail
            if op_to_on_rail.distance(op_to) > 15:
                op_to_ = op_to
            else:
                op_to_ = op_to_on_rail
            if (
                self.track_nr == 6012
                and self.name == 'StrUeb_6011_6012 - Berlin Gehrenseestraße'
                and km_from_db == '8,9 + 56'
            ):
                merge_from = "before"
            if (
                self.track_nr == 1280
                and self.name == 'StrUeb_1280_1284_3 - StrUeb_1255_1280_1'
                and km_from_db == '21,8 + 6'
            ):
                merge_from, op_from = None, op_from_on_rail
            if (
                self.track_nr == 5382
                and self.name == 'Ingolstadt Hbf - Ingolstadt-Seehof'
                and km_from_db == '-0,7 + -43'
            ):
                merge_from = None
            if (
                self.track_nr == 1760
                and self.name == 'Hannover Hbf - Hannover Bismarckstraße'
                and km_to_db == '1,0 + 2230'
            ):
                km_to_db = '1,1 + 2130'
            if (
                self.track_nr == 4011
                and self.name == 'Mannheim-Luzenberg - Mannheim-Waldhof'
                and km_to_db == '6,4 + 36'
            ):
                km_to_db = '6,4 +-442'
            if (
                self.track_nr == 2400
                and self.name == 'Düsseldorf-Rath Mitte - Düsseldorf-Rath'
                and km_to_db == '7,0 + 139'
            ):
                km_to_db = '7,1 + 39'
            if self.track_nr == 1750 and self.name == 'Seelze Mitte - Ahlem' and km_from_db == '11,5 + 0':
                merge_from = None
            if (
                self.track_nr == 2505
                and self.name == 'Rheinhausen Ost - Duisburg-Hochfeld Süd               (Hp)'
                and km_to_db == '13,8 + 257'
            ):
                km_to_db = '14.0 +    56'
            if self.track_nr == 6139 and self.name == 'StrUeb_6139_6170 - StrUeb_6078_6139_2':
                op_from_ = Point(3803586.416, 5828532.18)

            value_list_from = [km_from_db, [op_from_], merge_from]  # need to append geometric length
            value_list_to = [km_to_db, [op_to_], merge_after]  # need to append geometric length

            # add to dict
            track_km_dict[segment_id] += [value_list_from]
            track_km_dict[segment_id] += [value_list_to]

            # sort list of value_lists according to km values
            sorted_index = self.kilometrage.argsort_km_values([val_list[0] for val_list in track_km_dict[segment_id]])
            sorted_list = [track_km_dict[segment_id][i] for i in sorted_index]

            # drop lists where km==km_from or km==km_to (we use op_from and op_to for these km values)
            new_list = []
            dropped = []
            for i, value_list in enumerate(sorted_list):
                if len(value_list) == 3:
                    # always keep op_from/op_to
                    new_list.append(value_list)
                else:
                    # distance to other km points
                    # if any distance < 1
                    points = value_list[1]
                    point = points[0]
                    distances = [
                        point.distance(other) if point != other else np.inf
                        for other in [vlist[1][0] for vlist in sorted_list]
                    ]
                    c1 = any([d < 5 for d in distances])

                    # and both merge type None
                    nearest_point_idx = np.argmin(distances)
                    other = track_km_dict[segment_id][nearest_point_idx][1][0]
                    c2 = isinstance(value_list[2], type(None)) and isinstance(
                        sorted_list[nearest_point_idx][2], type(None)
                    )

                    # and both projections on km line yield same point
                    if self._n_km_lines == 2:
                        direction_idx = direction - 1
                    else:
                        direction_idx = direction
                    km_line_distances = [point.distance(km_line) for km_line in self.kilometrage.lines[direction_idx]]
                    km_line = self.kilometrage.lines[direction_idx][np.argmin(km_line_distances)]
                    point_projected = nearest_points(km_line, point)[0]
                    other_projected = nearest_points(km_line, other)[0]
                    c3 = point_projected.distance(other_projected) < 1e-3

                    prev_length = sorted_list[i - 1][-1]
                    if (
                        (c1 == True)
                        and (c2 == True)
                        and (c3 == True)
                        and (not (value_list[0] == km_from_db) and not (value_list[0] == km_to_db))
                        and prev_length != 0
                    ):
                        dropped.append(value_list)

                    elif not (value_list[0] == km_from_db) and not (value_list[0] == km_to_db):
                        new_list.append(value_list)

                    elif self.id == 27297 and (value_list[0] == km_from_db):
                        new_list.append(value_list)

                    else:
                        km_current = value_list[0]
                        length = value_list[-1]
                        if i < len(sorted_list) - 1:
                            for j in range(len(sorted_list[i + 1 :])):
                                if len(sorted_list[i + j + 1]) == 4:
                                    break
                            else:
                                j = None

                            if j:
                                km_next = sorted_list[i + j + 1][0]

                            else:
                                km_next = None

                        else:
                            km_next = None

                        dropped.append((km_current, km_next, length))

            sorted_list = new_list

            # adjust geometric length due to adding op_from and op_to
            for i, val_list in enumerate(sorted_list):
                if val_list[0] == km_from_db and len(val_list) == 3:
                    length = nominal_segment_length
                    val_list.append(length)
                    sorted_list[i] = val_list

                elif val_list[0] == km_to_db and len(val_list) == 3:
                    length = 0
                    val_list.append(length)
                    sorted_list[i] = val_list

            track_km_dict[segment_id] = sorted_list

        # integrate information in kilometrage
        for key in track_km_dict.keys():
            kms_value_list = deepcopy(track_km_dict[key])
            kms_lengths_list = deepcopy(track_length_dict[key])
            km_from = self.km_from_db
            km_to = self.km_to_db

            # sort km_value_list according to merge order
            merge_before_index = []
            merge_after_index = []
            for i, val_list in enumerate(kms_value_list):
                if val_list[2] == 'before':
                    merge_before_index.append(i)

                elif val_list[2] == 'after':
                    merge_after_index.append(i)

            remaining_index = [
                i for i in range(len(kms_value_list)) if not (i in merge_before_index) and not (i in merge_after_index)
            ]
            sorted_index = merge_before_index[::-1] + merge_after_index + remaining_index
            kms_value_list = [kms_value_list[i] for i in sorted_index]

            n_iter = 0
            pre_km_values = [l[0] for l in kms_value_list]
            pre_km_points = [l[1][0] for l in kms_value_list]
            self.kilometrage._set_pre_values(points=pre_km_points, values=pre_km_values)

            # delete double entries
            for i, values_list in enumerate(kms_value_list):
                n_iter += 1
                # dont add point if exact km already exists
                km_val_exists_list = [
                    values_list[0].replace(',', '.').replace(' ', '') == km_val.replace(',', '.').replace(' ', '')
                    for km_val in self.kilometrage.values_in_direction(direction=int(key[-1]))
                ]

                dir_key = 0 if int(key[-1]) < 2 else 1
                km_lines = self.kilometrage.lines[dir_key]
                km_line_0, km_line_end = km_lines[0], km_lines[-1]
                op_from_project = km_line_0.project(op_from, normalized=True)
                op_to_project = km_line_end.project(op_to, normalized=True)
                point = values_list[1][0]
                point_from_project = km_line_0.project(point, normalized=True)
                point_to_project = km_line_end.project(point, normalized=True)

                if any(km_val_exists_list):
                    continue

                # dont add point if km is larger than km_to (km_to: at target station of track segment)
                # excecpt: if point is to be merged with km line or km point is exactly end point of rail (for cases where op are not exactly at one end of a rail)
                if (
                    only_km_db(values_list[0]) > only_km_db(km_to)
                    and db_km_to_km(values_list[0]) > db_km_to_km(km_to)
                    and (point_to_project > op_to_project and point_to_project > 0 and point_to_project < 1)
                    and isinstance(values_list[2], type(None))
                ):
                    continue

                # dont add point if km is smaller than km_from (km_from: at source station of track segment)
                # excecpt: if point is to be merged with km line
                if (
                    only_km_db(values_list[0]) < only_km_db(km_from)
                    and db_km_to_km(values_list[0]) < db_km_to_km(km_from)
                    and (point_from_project > op_from_project and point_from_project > 0 and point_from_project < 1)
                    and isinstance(values_list[2], type(None))
                ):
                    continue

                if i == 0:
                    first_val = True
                else:
                    first_val = False

                if i == len(kms_value_list) - 1:
                    last_val = True
                else:
                    last_val = False

                self.kilometrage.integrate_kilometrage_points(
                    values_list=values_list,
                    direction=int(key[-1]),
                    boundary_kms=(km_from, km_to),
                    last_val=last_val,
                    first_val=first_val,
                )

            # reiterate to adjust nominal geomtric lengths
            # get lengths for original segments
            km_values = self.kilometrage.values_in_direction(direction=int(key[-1]))
            km_lengths = [[] for _ in range(len(km_values) - 1)]
            km_lengths_type = [[] for _ in range(len(km_values) - 1)]
            leftover_value_list = []

            for value_list in kms_lengths_list:
                km_1 = value_list[1]
                km_2 = value_list[2]
                length = value_list[3]

                for i, (km_s, km_e) in enumerate(zip(km_values[:-1], km_values[1:])):
                    first_km_match = (km_s == km_1) or (
                        km_s.replace(',', '.').replace(' ', '') == km_1.replace(',', '.').replace(' ', '')
                    )
                    second_km_match = (km_e == km_2) or (
                        km_e.replace(',', '.').replace(' ', '') == km_2.replace(',', '.').replace(' ', '')
                    )

                    if first_km_match == True and second_km_match == True:
                        km_lengths[i].append(length)
                        km_lengths_type[i].append('nominal')
                        break

                else:
                    leftover_value_list.append(value_list)

            # get lengths for segments that were splitted
            for value_list in leftover_value_list:
                km_1 = value_list[1]
                km_2 = value_list[2]
                length = value_list[3]

                for i, km in enumerate(km_values):
                    if km == km_1:
                        j = i
                        break
                else:
                    j = None
                for i, km in enumerate(km_values):
                    if km == km_2:
                        k = i
                        break
                else:
                    k = None

                if j and k:
                    assert k > j + 1, f'k {k} j {j}'
                    for i in range(k - j):
                        kmvals = km_values[j + i + 1 : k + 1]
                        km_lengths[j + i] = [
                            length
                            - np.sum(km_lengths[j : j + i])
                            - np.sum(
                                [1000 * (db_km_to_km(b) - db_km_to_km(a)) for (a, b) in zip(kmvals[:-1], kmvals[1:])]
                            )
                        ]
                        km_lengths_type[j + i] = ['nominal']

                elif j:
                    # difference between next km value and km_2
                    if j < len(km_values) - 1:
                        # km values in between
                        next_km_vals = [
                            km_values[j + i + 1]
                            for i in range(len(km_values[j + 1 :]))
                            if db_km_to_km(km_values[j + i]) < db_km_to_km(km_2)
                        ]

                        # difference to value after
                        difference_m = (db_km_to_km(next_km_vals[-1]) - db_km_to_km(km_2)) * 1000
                        # add to nominal length
                        length += difference_m

                        if len(next_km_vals) == 0:
                            pass

                        elif len(next_km_vals) == 1:
                            km_lengths[j] = [length]
                            km_lengths_type[j] = ['nominal']
                            if (
                                self.track_nr == 2273
                                and self.name == 'Maria-Veen - Coesfeld (Westf)'
                                and km_1 == '58,1 + 25'
                            ):
                                km_lengths_type[j] = ['computed']

                        else:
                            next_km_vals = [
                                km_values[j + i + 1]
                                for i in range(len(km_values[j + 1 :]))
                                if db_km_to_km(km_values[j + i]) < db_km_to_km(km_2)
                            ]
                            substractive = 0

                            for i in range(km_values.index(next_km_vals[-1]), j + 1, -1):
                                segment_length = (db_km_to_km(km_values[i]) - db_km_to_km(km_values[i - 1])) * 1000
                                substractive += segment_length
                                km_lengths[i - 1] = [segment_length]
                                km_lengths_type[i - 1] = ['computed']
                            km_lengths[j] = [length - substractive]
                            km_lengths_type[j] = ['nominal']

            for i, (a, b) in enumerate(zip(km_values[:-1], km_values[1:])):
                if len(km_lengths[i]) == 0:
                    km_lengths[i] = [1000 * (db_km_to_km(b) - db_km_to_km(a))]
                    km_lengths_type[i] = ['computed']

            lengths = flatten_list(km_lengths)
            lengths_type = flatten_list(km_lengths_type)
            assert len(lengths) == len(km_values) - 1, f'{lengths} | {km_values}'

            self.kilometrage._set_lengths_for_direction(lengths=lengths, direction=int(key[-1]))
            self.kilometrage._set_lengths_type_for_direction(lengths_type=lengths_type, direction=int(key[-1]))

        # recompute segment lengths
        self.kilometrage.compute_geometric_length()
        self.kilometrage.correct_point_positions()
        self.kilometrage.compute_geometric_length()
        self.kilometrage.compute_nominal_length()
        self.kilometrage.compute_geometric_offset()

    def _compute_kilometrage_offset(self):
        """
        Compute kilometrage offset based on differences between actual rail lengths (in ISR) and nominal rail lengths (from geo-streckennetz). This offset is needed to correct kilometrage for some track segments, because integrating kilometrage from DB Geo-Streckennetz dataset into ISR is hardly optimal, as kilometrage and track segment lengths can differ between the two. The offset computation is solely based on empirical testing.

        Offset is computed while considering kilometrage errors of previous track segments, which makes it necessary to create instances of previous track segments. This is done recursively, which is extremely inefficient due to having to preprocess each track segment multiple times. However, for now there is no other option, as fixing this would require much refactoring of the class.
        """
        name = self.name.split(' - ')[0]

        if self.has_additional_km_info == False or self._allow_previous == True:
            # get previous track segment
            previous_track_segment = self.from_track_and_name(
                track=self.track_nr,
                name=f'*{name}',
                to_km=self.properties['ISR_KM_VON'],
                allow_previous=False,
                enhance_kilometrage=self._enhance_kilometrage,
            )

            # if previous track segment exists
            if not isinstance(previous_track_segment, type(None)):
                # compute previous km error from previous track segment
                if previous_track_segment._n_km_lines == 1:
                    l = min(
                        [
                            len(previous_track_segment.kilometrage.lines[0]),
                            len(previous_track_segment.kilometrage.lengths_in_direction(direction=0)),
                        ]
                    )
                    prev_km_error_m = np.sum(
                        [
                            previous_track_segment.kilometrage.lines[0][j].length
                            - previous_track_segment.kilometrage.lengths_in_direction(direction=0)[j]
                            for j in range(l)
                        ]
                    )

                else:
                    l = min(
                        [
                            len(previous_track_segment.kilometrage.lines[0]),
                            len(previous_track_segment.kilometrage.lengths_in_direction(direction=1)),
                        ]
                    )
                    prev_km_error_m_1 = np.sum(
                        [
                            previous_track_segment.kilometrage.lines[0][j].length
                            - previous_track_segment.kilometrage.lengths_in_direction(direction=1)[j]
                            for j in range(l)
                        ]
                    )
                    l = min(
                        [
                            len(previous_track_segment.kilometrage.lines[1]),
                            len(previous_track_segment.kilometrage.lengths_in_direction(direction=2)),
                        ]
                    )
                    prev_km_error_m_2 = np.sum(
                        [
                            previous_track_segment.kilometrage.lines[1][j].length
                            - previous_track_segment.kilometrage.lengths_in_direction(direction=2)[j]
                            for j in range(l)
                        ]
                    )
                    prev_km_error_m = np.mean([prev_km_error_m_1, prev_km_error_m_2], dtype=float)
                self.kilometrage._previous_km_error = prev_km_error_m

                # compute km error
                if self._n_km_lines == 1:
                    l = min([len(self.kilometrage.lines[0]), len(self.kilometrage.lengths_in_direction(direction=0))])
                    km_error_m = np.sum(
                        [
                            self.kilometrage.lines[0][j].length - self.kilometrage.lengths_in_direction(direction=0)[j]
                            for j in range(l)
                        ]
                    )

                else:
                    l = min([len(self.kilometrage.lines[0]), len(self.kilometrage.lengths_in_direction(direction=1))])
                    km_error_m_1 = np.sum(
                        [
                            self.kilometrage.lines[0][j].length - self.kilometrage.lengths_in_direction(direction=1)[j]
                            for j in range(l)
                        ]
                    )
                    l = min([len(self.kilometrage.lines[1]), len(self.kilometrage.lengths_in_direction(direction=2))])
                    km_error_m_2 = np.sum(
                        [
                            self.kilometrage.lines[1][j].length - self.kilometrage.lengths_in_direction(direction=2)[j]
                            for j in range(l)
                        ]
                    )
                    km_error_m = np.mean([km_error_m_1, km_error_m_2], dtype=float)

                # compute errors and ratio
                rng = [0] if self._n_km_lines == 1 else [0, 1]
                prev_segment_error_m = np.mean(
                    [
                        previous_track_segment.kilometrage.lines[i][-1].length
                        - previous_track_segment.kilometrage.lengths[i][-1]
                        for i in rng
                    ],
                    dtype=float,
                )
                ratio = abs(km_error_m / prev_km_error_m)

                # compute kilometrage offset

                # Negative KM error is propageted from one track segment to later track segments
                if (prev_segment_error_m < 0 and self.has_additional_km_info == False) and km_error_m < 0:
                    # add previous segment error to offset
                    self.kilometrage._offset_from_start_m += prev_segment_error_m
                    # plus offset of previous segment, if previous segment had no km info and segment error < 0
                    if (not previous_track_segment.has_additional_km_info) and km_error_m < 0:
                        self.kilometrage._offset_from_start_m += previous_track_segment.kilometrage._offset_from_start_m

                # KM errors of previous segment and current segment negate each other (e.g. first segment is 100m to long, second segment is 100m to short)
                elif km_error_m * prev_km_error_m < 0 and ratio < 1.05 and ratio > 0.95:
                    self.kilometrage._offset_from_start_m += prev_km_error_m

                else:
                    # set offest to zero
                    self.kilometrage._offset_from_start_m = 0

            else:
                # set offest to zero
                self.kilometrage._offset_from_start_m = 0

        else:
            # set offest to zero
            self.kilometrage._offset_from_start_m = 0

    def _special_cases(self):
        """Handles certain special cases"""

        if self.name == 'Neuss Hbf - Weißenberg' and self.track_nr == 2610:
            self._km_to = km_to_db_km(db_km_to_km(self.km_to_db))  # overlength not correct in ISR
        elif self.name == 'Weißenberg - Meerbusch-Osterath' and self.track_nr == 2610:
            self._km_from = km_to_db_km(db_km_to_km(self.km_from_db))  # overlength not correct in ISR



import pyproj
from shapely.geometry.base import BaseGeometry
from shapely.ops import transform
from typing import Literal, TypeVar

# Type T: only instances of class BaseGeometry or subclasses (Point, Linestring, etc.)
T = TypeVar("T", bound="BaseGeometry")


class Transformer:
    """Initializes a Transformer object with the specified source and target coordinate systems.

    This class facilitates coordinate system transformations between the specified source and target coordinate systems.

    Attributes
    ----------
    _source : str
        The name of the source coordinate system.
    _target : str
        The name of the target coordinate system.
    _projection_from_source_to_target : callable
        A callable function for transforming coordinates from the source to the target coordinate system.
    _projection_from_target_to_source : callable
        A callable function for transforming coordinates from the target to the source coordinate system.
    """

    def __init__(self, source_cs: str, target_cs: str):
        """Initializes Transformer instance.

        Parameters
        ----------
        source_cs : str
            The source coordinate system as EPSG code.
        target_cs : str
            The target coordinate system as EPSG code."""
        # check if cs exist
        # TODO

        # store coordinate system names
        self._source = source_cs
        self._target = target_cs

        # define projection from source to target
        self._projection_from_source_to_target = pyproj.Transformer.from_proj(
            pyproj.Proj(source_cs), pyproj.Proj(target_cs)
        ).transform

        # define projection from target to source
        self._projection_from_target_to_source = pyproj.Transformer.from_proj(
            pyproj.Proj(target_cs), pyproj.Proj(source_cs)
        ).transform

    # define attributes as read-only
    @property
    def projection_from_source_to_target(self):
        return self._projection_from_source_to_target

    @property
    def projection_from_target_to_source(self):
        return self._projection_from_target_to_source

    @property
    def source(self):
        return self._source

    @property
    def target(self):
        return self._target

    def transform(self, projection: Literal["source_to_target", "target_to_source"], object: T) -> T:
        """Transforms a shapely.Geometry object from a source coordinate system to a target coordinate system.

        Parameters
        ----------
        projection: Literal["source_to_target", "target_to_source"]
            Defines whether to transform from source coordinate system to target coordinate system or the other way around.
        object: BaseGeometry
            The object to transform. The type of object must be a subclass of shapely.Geometry, e.g Point, LineString, MultiLineString, etc.

        Returns
        -------
        object: BaseGeometry
            A new instance of the same object class with transformed coordinates.

        Note
        ----
        Consider order of axes when transforming. For WGS84, axes order is (latitude, longitude).
        """

        # check type of object
        if not issubclass(type(object), BaseGeometry):
            raise ValueError(
                f"Unknown type of object: {type(object)}. Object must be a subclass of shapely.Geometry, e.g. Point, LineString, MulitLineString, etc."
            )

        if projection == "source_to_target":
            # project from source to target cs
            transformed_object = transform(self.projection_from_source_to_target, object)

        elif projection == "target_to_source":
            # project from target to source cs
            transformed_object = transform(self.projection_from_target_to_source, object)

        else:
            # raise error: unknown projection
            raise ValueError(
                f"Unkown projection: {projection}. Possible values are 'source_to_target' and 'target_to_source'."
            )

        return transformed_object



from __future__ import annotations
from shapely import LineString, Point
from typing import Tuple, Annotated, TYPE_CHECKING
from isr_matcher.data_handlers.transformer import Transformer

# from numpy.typing import ArrayLike
# from pyrailmapping._constants._properties import ISR_PROPERTIES_TRACK_SEGMENTS
# from pyrailmapping._constants._parse import dtype
# import numpy as np

# false at run time (for type hint only)
# if TYPE_CHECKING:
#    from pyrailmapping.geometry.operational_point import OperationalPoint
#    from pyrailmapping.geometry.track_segment import TrackSegment


class Rail(LineString):
    """Represents a rail. Inherits from LineString so shapely functions can be used on instances directly.

    Attributes
    ----------
    _id_to_attrs : dict
        Dictionary mapping ID to attributes.
    __slots__ : tuple
        Slots for assigning __class__.
    """

    _id_to_attrs = {}
    __slots__ = (
        LineString.__slots__
    )  # slots must be the same for assigning __class__ - https://stackoverflow.com/a/52140968

    def __init__(
        self, coords: Tuple[float, float], lines_c: list, layout: dict, name: str, track: int, index: int | None = None
    ):
        """Initializes a Rail object.

        Parameters
        ----------
        coords : Tuple[float, float]
            Coordinates of the rail.
        lines_c : list
            List of lines.
        layout : dict
            Dictionary containing layout information.
        name : str
            Name of the rail.
        track : int
            Track number.
        index : int, optional
            Index of the rail, by default None
        """
        # get lines in direction from track segment
        lines = [LineString(coords) for coords in lines_c]
        lines_1 = [lines[i] for i in range(len(lines)) if i in layout['direction_1_rails']]
        lines_2 = [lines[i] for i in range(len(lines)) if i in layout['direction_2_rails']]

        # determine direction for rail
        line = LineString(coords)
        if line in lines_1 and line in lines_2:  # monorails (could be directed but treated as bi-directional)
            direction = 0
        elif line in lines_1:
            direction = 1
        elif line in lines_2:
            direction = 2
        else:  # sidings (could be directed but treated as bi-directional)
            direction = 0

        self._id_to_attrs[id(self)] = dict(
            lines=lines, layout=layout, name=name, direction=direction, track=track, index=index
        )

    def __new__(
        cls, coords: Tuple[float, float], lines_c: list, layout: dict, name: str, track: int, index: int | None = None
    ) -> Rail:
        """Creates a new instance of Rail.

        Parameters
        ----------
        coords : Tuple[float, float]
            Coordinates of the rail.
        lines_c : list
            List of lines.
        layout : dict
            Dictionary containing layout information.
        name : str
            Name of the rail.
        track : int
            Track number.
        index : int, optional
            Index of the rail, by default None

        Returns
        -------
        Rail
            A new instance of Rail.
        """
        line_string = super().__new__(cls, coords)
        line_string.__class__ = cls
        return line_string

    @property
    def index(self) -> int | None:
        """Returns index."""
        return self._id_to_attrs[id(self)]['index']

    @index.setter
    def index(self, value: int | None):
        """Sets index of rail."""
        self._id_to_attrs[id(self)]['index'] = value

    @property
    def direction(self) -> int:
        """Returns direction of rail."""
        return self._id_to_attrs[id(self)]['direction']

    @direction.setter
    def direction(self, value: int):
        """Sets direction for rail."""
        self._id_to_attrs[id(self)]['direction'] = value

    @property
    def track(self) -> str:
        """Returns track number the rail belongs to."""
        return self._id_to_attrs[id(self)]['track']

    @track.setter
    def track(self, value: int):
        """Sets the track number."""
        self._id_to_attrs[id(self)]['track'] = value

    @property
    def track_segment_name(self) -> str:
        """Returns name of track segment the rail belongs to."""
        return self._id_to_attrs[id(self)]['name']

    @track_segment_name.setter
    def track_segment_name(self, value: str):
        """Sets the track segment name."""
        self._id_to_attrs[id(self)]['name'] = value

    @property
    def layout(self) -> dict:
        """Returns layout of the track segmetn the rail belongs to."""
        return self._id_to_attrs[id(self)]['layout']

    @layout.setter
    def layout(self, value: dict):
        """Sets the track segment layout."""
        self._id_to_attrs[id(self)]['layout'] = value

    @property
    def lines(self) -> list[LineString]:
        """Returns the all rails of the track segment the line belongs to."""
        return self._id_to_attrs[id(self)]['lines']

    @lines.setter
    def lines(self, value: list):
        """Sets the track segment line coordinates."""
        self._id_to_attrs[id(self)]['lines'] = value

    @property
    def coords_wgs(self) -> list[Point]:
        transformer = Transformer(source_cs='epsg:31467', target_cs='epsg:4326')
        points_wgs = []
        for coord_utm in self.coords[:]:
            coord_wgs = transformer.transform('source_to_target', Point([coord_utm[1], coord_utm[0]]))
            points_wgs.append(Point(coord_wgs))
        return points_wgs

    @classmethod
    def from_linestring_and_rail(cls, rail: Rail, line_string: LineString) -> Rail:
        """Creates a new rail instance with coordinates from 'line_string' and all other attributes copied from 'rail'.

        Parameters
        ----------
        other: Rail
            An instance of class Rail whose attributes will be copied to self.
        """
        new_rail = cls(
            coords=line_string.coords,  # type: ignore
            lines_c=rail.lines,
            layout=rail.layout,
            name=rail.track_segment_name,
            track=int(rail.track),
            index=rail.index,
        )

        new_rail.direction = rail.direction

        return new_rail

    # Pickling support
    def __reduce__(self):
        return (
            self.__class__,
            (
                tuple(self.coords),
                [line.coords for line in self.lines],
                self._id_to_attrs[id(self)]['layout'],
                self._id_to_attrs[id(self)]['name'],
                self._id_to_attrs[id(self)]['track'],
                self._id_to_attrs[id(self)]['index'],
            ),
        )



from __future__ import annotations
from pathlib import Path
from typing import Literal, Union, Tuple, TYPE_CHECKING
from numpy.typing import ArrayLike
import numpy as np
from shapely import Point, LineString
from isr_matcher.data_handlers.transformer import Transformer
from isr_matcher.ops._functions import angle_between
from isr_matcher._constants._parse import FLOAT_TYPES
import datetime


class GNSSSeries:

    """
    Container class for GPS series.

    Attributes
    ----------
    line_wgs: LineString
        Shapely LineString of measurements in WGS84 reference system (EPSG:4326). Axes: longitude, latitude.
    line_utm: LineString
        Shapely LineString of measurements in UTM reference system (EPSG:31467). Axes: x, y.
    time: np.ndarray
        Array of datetimes.
    time_s: np.ndarray
        Array of relative times from first time stamp, in seconds.
    latitude: np.ndarray
        Array of latitude coordinates (WGS84 / EPSG:4326).
    longitude: np.ndarray
        Array of longitude coordinates (WGS84 / EPSG:4326).
    x: np.ndarray
        Array of horizontal coordinates (UTM / EPSG:31467).
    y: np.ndarray
        Array of vertical coordinates (UTM / EPSG:31467).
    altitude_m: np.ndarray | None
        Array of altitudes in m, if available. Else, None.
    veolicty_ms: np.ndarray | None
        Array of velocities in m/s, if available. Else, None.
    acceleration_ms2: np.ndarray | None
        Array of acceleartions in m/s^2, if available. Else, None.
    """

    def __init__(
        self,
        time_utc: list[datetime.datetime] | np.ndarray | None,
        time_s: list[float] | np.ndarray | None,  # in seconds!
        latitude: list[float] | np.ndarray,  # in WGS84!
        longitude: list[float] | np.ndarray,  # in WGS84!
        altitude_m: list[float] | np.ndarray | None = None,
        velocity_ms: list[float] | np.ndarray | None = None,
        acceleration_ms2: list[float] | np.ndarray | None = None,
    ):
        # check_input
        self._check_input(time_utc, time_s, latitude, longitude, altitude_m, velocity_ms, acceleration_ms2)

        # project from wgs84 to utm
        transformer = Transformer(source_cs='epsg:4326', target_cs='epsg:31467')  # set up Transformer
        gnss_np = np.column_stack(tup=(latitude, longitude))  # gnss array (lat, lon)
        gnss_line_string_wgs_latlon = LineString(gnss_np)  # linestring
        gnss_line_string_utm_yx = transformer.transform(  # transform
            projection='source_to_target', object=gnss_line_string_wgs_latlon
        )
        gnss_line_string_utm_xy = LineString(  # invert axes
            [(y, x) for (x, y) in zip(gnss_line_string_utm_yx.xy[0], gnss_line_string_utm_yx.xy[1])]
        )

        # time in seconds
        if not self.notnone(time_s):
            time_s = np.array([(t - time_utc[0]) / np.timedelta64(1, 's') for t in time_utc])  # type: ignore
            if any(np.diff(time_s) == 0):
                idx = np.where(np.diff(time_s) == 0)[0]
                for i in idx:
                    # mean time diff
                    time_diff = np.median(np.diff(time_s))

                    # i and i+1 same time
                    time_s[i + 1 :] = time_s[i + 1 :] + time_diff

        if not self.notnone(time_utc):
            time_utc = np.ones(len(time_s)) * np.nan

        # set attributes
        self._line_wgs = LineString(np.column_stack(tup=(longitude, latitude)))
        self._line_utm = gnss_line_string_utm_xy
        self._time = time_utc
        self._time_s = time_s
        self._latitude = np.array(latitude)
        self._longitude = np.array(longitude)
        self._x = np.array(gnss_line_string_utm_xy.xy[0])
        self._y = np.array(gnss_line_string_utm_xy.xy[1])
        self._altitude_m = np.array(altitude_m) if not isinstance(altitude_m, type(None)) else None
        self._velocity_ms = np.array(velocity_ms) if not isinstance(velocity_ms, type(None)) else None
        self._acceleration_ms2 = np.array(acceleration_ms2) if not isinstance(acceleration_ms2, type(None)) else None
        self._sigma = None
        self._error = None

    # getter methods
    @property
    def line_wgs(self) -> LineString:
        """Returns a LineString of WGS84 coordinates. Axes: (longitude, latitude).

        Returns
        -------
        line_wgs: LineString
            LineString of WGS84 coordinates. Axes: (longitude, latitude).
        """
        return self._line_wgs

    @property
    def line_utm(self) -> LineString:
        """Returns a LineString of UTM (EPSG:31467) coordinates. Axes: (horizontal, vertical).

        Returns
        -------
        line_utm: LineString
            LineString of UTM (EPSG:31467) coordinates. Axes: (horizontal, vertical).
        """
        return self._line_utm

    @property
    def coords_wgs(self) -> list[Point]:
        """Returns a list of GNSS coordinates as Points (WGS84). Axes: (longitude, latitude).

        Returns
        -------
        coords_wgs: LineString
            List of WGS84 coordinates. Axes: (longitude, latitude).
        """
        return [Point(coord) for coord in self.line_wgs.coords]

    @property
    def coords_utm(self) -> list[Point]:
        """Returns a list of GNSS coordinates as Points (UTM: EPSG:31467). Axes: (horizontal, vertical).

        Returns
        -------
        coords_utm: LineString
            List of UTM coordinates (EPSG:31467). Axes: (horizontal, vertical).
        """
        return [Point(coord) for coord in self.line_utm.coords]

    @property
    def time(self) -> np.ndarray:
        """Returns an array of time stamps in datetime format.

        Returns
        -------
        time: np.ndarray
            Array of time stamps in datetime format.
        """
        return self._time

    @property
    def time_s(self) -> np.ndarray:
        """Returns an array of relative time differences from first time stamp, in seconds.

        Returns
        -------
        time_s: np.ndarray
            Array of relative time differences from first time stamp, in seconds.
        """
        return self._time_s

    @property
    def longitude(self) -> np.ndarray:
        """Returns an array of longitudes (WGS84).

        Returns
        -------
        longitude: np.ndarray
            Array of longitudes (WGS84).
        """
        return self._longitude

    @property
    def latitude(self) -> np.ndarray:
        """Returns an array of latitudes

        Returns
        -------
        latitude: np.ndarray
            Array of latitude (WGS84).
        """
        return self._latitude

    @property
    def x(self) -> np.ndarray:
        """Returns an array of hoizontal UTM coordinates (EPSG:31467).

        Returns
        -------
        x: np.ndarray
            Array of hoizontal UTM coordinates (EPSG:31467).
        """
        return self._x  # type: ignore

    @property
    def y(self) -> np.ndarray:
        """Returns an array of vertical UTM coordinates (EPSG:31467).

        Returns
        -------
        y: np.ndarray
            Array of vertical UTM coordinates (EPSG:31467).
        """
        return self._y  # type: ignore

    @property
    def altitude_m(self) -> np.ndarray | None:
        """Returns an array of altitudes in m, if available. Else returns None.

        Returns
        -------
        altitude_m: np.ndarray
            Array of altitudes in m, if available. Else None.
        """
        return self._altitude_m

    @property
    def velocity_ms(self) -> np.ndarray | None:
        """Returns an array of velocities in m/s, if available. Else returns None.

        Returns
        -------
        velocity_ms: np.ndarray
            Array of velocities in m/s, if available. Else None.
        """
        return self._velocity_ms

    @property
    def acceleration_ms2(self) -> np.ndarray | None:
        """Returns an array of accelerations in m/s^2, if available. Else returns None.

        Returns
        -------
        acceleration_ms2: np.ndarray
            Array of accelerations in m/s^2, if available. Else None.
        """
        return self._acceleration_ms2

    @property
    def sigma(self) -> float | None:
        """Returns sigma (Estimate of GNSS noise standard deviation), if available. Else returns None.

        Returns
        -------
        sigma: float | None
            Estimate of GNSS noise standard deviation. If not set, returns None.
        """
        return self._sigma

    @sigma.setter
    def sigma(self, value: float):
        """Set sigma (Estimate of GNSS noise standard deviation)."""
        self._sigma = value

    @property
    def error(self) -> float | None:
        """Returns gnss error (Estimate of GNSS error at each time step), if available. Else returns None.

        Returns
        -------
        error: float | None
            Estimate of GNSS error at each time step. If not set, returns None.
        """
        return self._sigma

    @error.setter
    def error(self, value: np.ndarray):
        """Set error (Estimate of GNSS error at each time step)."""
        self._error = value

    def __len__(self):
        """Method for calling len() on instances."""
        return len(self.x)

    def __getitem__(self, key):
        """Enables selecting subset of GNSSSeries by slice indexing"""
        if isinstance(key, slice):
            # Get the start, stop, and step from the slice
            start, stop, step = key.indices(len(self))
            # check for None attributes
            if not isinstance(self.altitude_m, type(None)):
                sliced_altitude_m = self.altitude_m[start:stop:step]
            else:
                sliced_altitude_m = None
            if not isinstance(self.velocity_ms, type(None)):
                sliced_velocity_ms = self.velocity_ms[start:stop:step]
            else:
                sliced_velocity_ms = None
            if not isinstance(self.acceleration_ms2, type(None)):
                sliced_acceleration_ms2 = self.acceleration_ms2[start:stop:step]
            else:
                sliced_acceleration_ms2 = None

            return GNSSSeries(
                time=self.time[start:stop:step],
                latitude=self.latitude[start:stop:step],
                longitude=self.longitude[start:stop:step],
                altitude_m=sliced_altitude_m,
                velocity_ms=sliced_velocity_ms,
                acceleration_ms2=sliced_acceleration_ms2,
            )
        else:  # single indexing not supported
            raise TypeError("Invalid argument type.")

    def prune(self, r: float, index: list | None = None):
        """Prunes noisy measurments within r of previous measurement. If index is given, prune by index."""
        temp = False
        if isinstance(index, type(None)):
            delete_index = []
            temp_delete_index = []
            last_point = Point(self.x[0], self.y[0])
            angle = 0.0
            for i in range(len(self) - 1):
                # distance between neighbor points
                next_point = Point(self.x[i + 1], self.y[i + 1])
                distance = last_point.distance(next_point)

                # angle
                if last_point != Point(self.x[0], self.y[0]):
                    p0 = np.array([second_last_point.x, second_last_point.y])
                    p1 = np.array([last_point.x, last_point.y])
                    p2 = np.array([next_point.x, next_point.y])
                    angle = angle_between(v1=p1 - p0, v2=p2 - p1)

                # compare with standard deviation and angle threshold
                if distance < r:
                    delete_index.append(i + 1)
                elif angle > 100:
                    temp_delete_index.append(i + 1)
                    temp = True
                else:
                    second_last_point = last_point
                    last_point = next_point

                if len(temp_delete_index) > 10 and np.all(np.diff(temp_delete_index) == 0):
                    temp_delete_index = []
                    second_last_point = last_point
                    last_point = next_point
                    temp = False
                elif len(temp_delete_index) > 10:
                    delete_index = np.unique(delete_index + temp_delete_index).tolist()
                    temp_delete_index = []
                    second_last_point = last_point
                    last_point = next_point
                    temp = False

        else:
            delete_index = index

        # prune noisy measurements
        if temp == True:
            delete_index = np.unique(delete_index + temp_delete_index).tolist()
        self._time = np.delete(self._time, delete_index)
        self._time_s = np.delete(self._time_s, delete_index)
        self._latitude = np.delete(self._latitude, delete_index)
        self._longitude = np.delete(self._longitude, delete_index)
        self._x = np.delete(self._x, delete_index)
        self._y = np.delete(self._y, delete_index)
        self._velocity_ms = (
            np.delete(self._velocity_ms, delete_index) if not isinstance(self._velocity_ms, type(None)) else None
        )
        self._altitude_m = (
            np.delete(self._altitude_m, delete_index) if not isinstance(self._altitude_m, type(None)) else None
        )
        self._acceleration_ms2 = (
            np.delete(self._acceleration_ms2, delete_index)
            if not isinstance(self._acceleration_ms2, type(None))
            else None
        )
        self._line_wgs = LineString(np.column_stack(tup=(self._longitude, self._latitude)))
        self._line_utm = LineString(np.column_stack(tup=(self._x, self._y)))

    @staticmethod
    def average_by_index(array, index):
        """Replaces consecutive values of an array given by index with their average."""
        return np.array([np.mean(array[idx].astype(float)) for idx in index], dtype=array.dtype)

    def average_low_velocity(self, threshold: float = 3.0):
        """Averages sequences in time series where velocity is smaller than 'threshold', in meter per second."""

        assert not isinstance(self.velocity_ms, type(None)), 'velocity_ms is None'

        # find consecutive sequences where all values are smaller than threshold
        mask = self.velocity_ms < threshold
        split_indices = np.where(~mask)[0]

        # split the array indices
        sequences_index = np.split(np.arange(len(self.velocity_ms)), split_indices)
        sequences_index = [arr for arr in sequences_index if len(arr) > 0]

        # average sequences for all parameters
        self._time = self.average_by_index(
            self._time, sequences_index
        )  # np.array([self._time[idx][0] for idx in sequences_index])
        self._time_s = self.average_by_index(self._time_s, sequences_index)
        self._latitude = self.average_by_index(self._latitude, sequences_index)
        self._longitude = self.average_by_index(self._longitude, sequences_index)
        self._x = self.average_by_index(self._x, sequences_index)
        self._y = self.average_by_index(self._y, sequences_index)
        self._velocity_ms = (
            self.average_by_index(self._velocity_ms, sequences_index)
            if not isinstance(self._velocity_ms, type(None))
            else None
        )
        self._altitude_m = (
            self.average_by_index(self._altitude_m, sequences_index)
            if not isinstance(self._altitude_m, type(None))
            else None
        )
        self._acceleration_ms2 = (
            self.average_by_index(self._acceleration_ms2, sequences_index)
            if not isinstance(self._acceleration_ms2, type(None))
            else None
        )
        self._line_wgs = LineString(np.column_stack(tup=(self._longitude, self._latitude)))
        self._line_utm = LineString(np.column_stack(tup=(self._x, self._y)))

    def boundary_string(self, r: float = 100.0) -> str:
        """Computes a boundary with radius 'r' around the gnss series and returns a string of WGS84 coordinates which can be used as input when querying with filter 'BOUNDARY' (see RailDataImport).

        Parameters
        ----------
        r: float
            The radius of the boundary, in meter. Default: 100.

        Returns
        -------
        coord_string: str
            String of WGS84 coordinates representing the boundary.
        """

        buffer = self.line_utm.buffer(r)  # compute buffer
        x, y = buffer.exterior.xy  # get coordinate arrays
        coord_string = ''.join([f'{x} {y} ' for x, y in zip(x, y)])  # create string
        return coord_string

    def notnone(self, value) -> bool:
        if not isinstance(value, type(None)):
            return True
        else:
            return False

    def _check_input(
        self,
        time_utc: list[datetime.datetime] | np.ndarray | None,
        time_s: list[float] | np.ndarray | None,
        latitude: list[float] | np.ndarray,  # in WGS84!
        longitude: list[float] | np.ndarray,  # in WGS84!
        altitude_m: list[float] | np.ndarray | None = None,
        velocity_ms: list[float] | np.ndarray | None = None,
        acceleration_ms2: list[float] | np.ndarray | None = None,
    ):
        if self.notnone(time_utc) and self.notnone(time_s):
            time = time_utc
        elif self.notnone(time_utc):
            time = time_utc
        elif self.notnone(time_s):
            time = time_s
        else:
            raise ValueError("Either 'time_utc' or 'time_s' must be given. Both are None")

        if len(time) == 0:
            raise ValueError("Input list with zero length.")

        # check input length
        arrays_equal_length = len(time) == len(latitude) == len(longitude)
        if not isinstance(altitude_m, type(None)):
            arrays_equal_length = arrays_equal_length and (len(time) == len(altitude_m))
        if not isinstance(velocity_ms, type(None)):
            arrays_equal_length = arrays_equal_length and (len(time) == len(velocity_ms))
        if not isinstance(acceleration_ms2, type(None)):
            arrays_equal_length = arrays_equal_length and (len(time) == len(acceleration_ms2))
        if not arrays_equal_length:
            raise ValueError(f"All input parameters must have the same length.")

        # check time
        if not isinstance(time, (list, np.ndarray)):
            raise TypeError(
                f"Parameter 'time' must be of type <list> or <numpy.ndarray>. Passed type was {type(time)}."
            )

        else:
            if not all([isinstance(val, np.datetime64) or isinstance(val, float) for val in time]):
                raise TypeError(
                    "All elements in 'time_utc' must be of type <numpy.datetime64>, all elements in 'time_s' must be of type float."
                )

        # check latitude
        if not isinstance(latitude, (list, np.ndarray)):
            raise TypeError(
                f"Parameter 'latitude' must be of type <list> or <numpy.ndarray>. Passed type was {type(latitude)}."
            )

        else:
            if not all([isinstance(val, FLOAT_TYPES) for val in latitude]):  # type: ignore
                raise TypeError("All elements in 'latitude' must be of type <float>.")

        # check longitude
        if not isinstance(longitude, (list, np.ndarray)):
            raise TypeError(
                f"Parameter 'longitude' must be of type <list> or <numpy.ndarray>. Passed type was {type(longitude)}."
            )

        else:
            if not all([isinstance(val, FLOAT_TYPES) for val in longitude]):  # type: ignore
                raise TypeError("All elements in 'longitude' must be of type <float>.")

        # check altitude
        if not isinstance(altitude_m, type(None)):
            if not isinstance(altitude_m, (list, np.ndarray)):
                raise TypeError(
                    f"Parameter 'altitude' must be of type <list> or <numpy.ndarray>. Passed type was {type(altitude_m)}."
                )

            else:
                if not all([isinstance(val, FLOAT_TYPES) for val in altitude_m]):  # type: ignore
                    raise TypeError("All elements in 'altitude' must be of type <float>.")

        # check velocity_ms
        if not isinstance(velocity_ms, type(None)):
            if not isinstance(velocity_ms, (list, np.ndarray)):
                raise TypeError(
                    f"Parameter 'velocity_ms' must be of type <list> or <numpy.ndarray>. Passed type was {type(velocity_ms)}."
                )

            else:
                if not all([isinstance(val, FLOAT_TYPES) for val in velocity_ms]):  # type: ignore
                    raise TypeError("All elements in 'velocity_ms' must be of type <float>.")

        # check acceleration_ms2
        if not isinstance(acceleration_ms2, type(None)):
            if not isinstance(acceleration_ms2, (list, np.ndarray)):
                raise TypeError(
                    f"Parameter 'acceleration_ms2' must be of type <list> or <numpy.ndarray>. Passed type was {type(acceleration_ms2)}."
                )

            else:
                if not all([isinstance(val, FLOAT_TYPES) for val in acceleration_ms2]):  # type: ignore
                    raise TypeError("All elements in 'acceleration_ms2' must be of type <float>.")
